import gettext
import json
import locale
import logging
import os
from typing import Dict, Union

from eduvpn.settings import COUNTRY, COUNTRY_MAP, LANGUAGE
from eduvpn.utils import get_prefix
from eduvpn.variants import ApplicationVariant

logger = logging.getLogger(__name__)

country_mapping = None


def initialize(app_variant: ApplicationVariant):
    prefix = get_prefix()
    try:
        setup(app_variant, prefix)
        logger.debug("i18n successfully initialized")
    except Exception as e:
        logger.error(f"i18n initialization failed: {e}")


def setup(app_variant: ApplicationVariant, prefix: str):
    """
    Init locale and gettext, returns text domain
    """
    domain = app_variant.translation_domain
    directory = os.path.join(prefix, "share/locale")

    locale.setlocale(locale.LC_ALL, "")
    locale.bindtextdomain(domain, directory)  # type: ignore
    locale.textdomain(domain)  # type: ignore
    gettext.bindtextdomain(domain, directory)
    gettext.textdomain(domain)

    return domain


def country() -> str:
    try:
        locale_setting = locale.getlocale()[0]
        if not locale_setting:
            return COUNTRY
        return locale_setting.replace("_", "-")
    except Exception:
        return COUNTRY


def language() -> str:
    try:
        locale_setting = locale.getlocale()[0]
        if not locale_setting:
            return LANGUAGE
        return locale_setting.split("_")[0]
    except Exception:
        return LANGUAGE


def extract_translation(d: Union[str, Dict[str, str]]):
    if isinstance(d, dict):
        for m in [country(), language(), "en-US", "en"]:
            try:
                return d[m]
            except KeyError:
                continue
        return list(d.values())[0]  # otherwise just return first in list
    else:
        return d


def retrieve_country_name(country_code: str) -> str:
    country_map = _read_country_map()
    if country_code not in country_map:
        return country_code
    return extract_translation(country_map[country_code])


def _read_country_map() -> dict:
    """
    Read the storage from disk, returns an empty dict in case of failure.
    """
    global country_mapping

    if country_mapping is not None:
        return country_mapping

    if COUNTRY_MAP.exists():
        try:
            with open(COUNTRY_MAP, "r") as f:
                country_mapping = json.load(f)
                return country_mapping
        except Exception as e:
            logger.error(f"Error reading country map: {e}")
    return {}



import json
from configparser import ConfigParser
from datetime import datetime, timedelta
from enum import IntEnum
from typing import Any, Dict, List, Optional


class Token:
    """The class that represents oauth Tokens
    :param: access: str: The access token
    :param: refresh: str: The refresh token
    :param: expired: int: The expire unix time
    """

    def __init__(self, access: str, refresh: str, expired: int):
        self.access = access
        self.refresh = refresh
        self.expires = expired

    def dump(self) -> str:
        """Dumps the tokens as a JSON string"""
        d = {
            "access_token": self.access,
            "refresh_token": self.refresh,
            "expires": self.expires,
        }

        return json.dumps(d)


class Protocol(IntEnum):
    UNKNOWN = 0
    OPENVPN = 1
    WIREGUARD = 2
    WIREGUARDTCP = 3


class Config:
    """The class that represents an OpenVPN/WireGuard config
    :param: config: str: The config string
    :param: protocol: Protocol: The type of config, openvpn/wireguard
    :param: default_gateway: bool: If this configuration should be configured with default gateway
    """

    def __init__(
        self,
        config: str,
        protocol: Protocol,
        default_gateway: bool,
        dns_search_domains: List[str],
        should_failover: bool,
    ):
        self.config = config
        self.protocol = protocol
        self.default_gateway = default_gateway
        self.dns_search_domains = dns_search_domains
        self.should_failover = should_failover

    def __str__(self):
        return self.config


def parse_tokens(tokens_json: str) -> Token:
    jsonT = json.loads(tokens_json)
    return Token(jsonT["access_token"], jsonT["refresh_token"], jsonT["expires_at"])


def parse_config(config_json: str) -> Config:
    d = json.loads(config_json)
    cfg = Config(
        d["config"],
        Protocol(d["protocol"]),
        d["default_gateway"],
        d.get("dns_search_domains", []),
        d["should_failover"],
    )
    return cfg


class Validity:
    def __init__(
        self,
        start: datetime,
        end: datetime,
        button: datetime,
        countdown: datetime,
        notifications: List[datetime],
    ):
        self.start = start
        self.end = end
        self.button = button
        self.countdown = countdown
        self.notifications = notifications

    @property
    def remaining(self) -> timedelta:
        """
        Return the duration from now until expiry.
        """
        return self.end - datetime.now()

    @property
    def is_expired(self) -> bool:
        """
        Return True if the validity has expired.
        """
        return datetime.now() >= self.end


def parse_date(d: Dict[str, Any], key: str) -> datetime:
    val = d.get(key, 0)
    return datetime.fromtimestamp(val)


def parse_expiry(exp_json: str) -> Optional[Validity]:
    d = json.loads(exp_json)
    start = parse_date(d, "start_time")
    end = parse_date(d, "end_time")
    button = parse_date(d, "button_time")
    countdown = parse_date(d, "countdown_time")
    notifs = d.get("notification_times", [])
    parsed_notifs = []
    for n in notifs:
        parsed_notifs.append(datetime.fromtimestamp(n))
    return Validity(start, end, button, countdown, parsed_notifs)


class Connection:
    "Base class for connection configurations."

    @classmethod
    def parse(cls, config: Config) -> "Connection":
        if config.protocol == Protocol.WIREGUARD or config.protocol == Protocol.WIREGUARDTCP:
            connection_type = WireGuardConnection
        else:
            connection_type = OpenVPNConnection  # type: ignore
        return connection_type.parse(config.config)

    def connect(self, callback):
        """
        Start this connection.

        This method is always called from the main thread.
        """
        raise NotImplementedError


class OpenVPNConnection(Connection):
    def __init__(self, config_str):
        self.config_str = config_str
        super().__init__()

    @classmethod
    def parse(cls, config_str: str) -> "OpenVPNConnection":  # type: ignore
        return cls(config_str=config_str)

    def connect(
        self,
        manager,
        default_gateway,
        allow_lan,
        dns_search_domains,
        proxy,
        callback,
    ):
        manager.start_openvpn_connection(
            self.config_str,
            default_gateway,
            dns_search_domains,
            callback=callback,
        )


class WireGuardConnection(Connection):
    def __init__(self, config: ConfigParser):
        self.config = config
        super().__init__()

    @classmethod
    def parse(cls, config_str: str) -> "WireGuardConnection":  # type: ignore
        config = ConfigParser()
        config.read_string(config_str)
        return cls(config=config)

    @property
    def proxy_endpoint(self) -> Optional[str]:
        return self.config["Peer"].get("ProxyEndpoint", None)

    def connect(
        self,
        manager,
        default_gateway,
        allow_lan,
        dns_search_domains,
        proxy,
        callback,
    ):
        manager.start_wireguard_connection(
            self.config,
            default_gateway,
            allow_wg_lan=allow_lan,
            callback=callback,
            proxy=proxy,
        )



from contextlib import closing
from socket import AF_INET, SOCK_DGRAM, SOCK_STREAM, socket
from typing import List
from urllib.parse import urlparse

from eduvpn.utils import handle_exception


def find_free_udp_port():
    with closing(socket(AF_INET, SOCK_DGRAM)) as s:
        s.bind(("", 0))
        port = s.getsockname()[1]
        return port


def find_free_tcp_port():
    with closing(socket(AF_INET, SOCK_STREAM)) as s:
        s.bind(("", 0))
        port = s.getsockname()[1]
        return port


class Proxy:
    """The class that represents a proxyguard instance
    :param: common: The common library
    :param: peer: str: The remote peer string
    :param: listen: str: The listen proxy string
    """

    def __init__(
        self,
        common,
        endpoint,
    ):
        self.common = common
        self.endpoint = endpoint
        self.wrapper = None
        self._cached_source_port = None
        self._cached_listen_port = None

    def forward_exception(self, error):
        handle_exception(self.common, error)

    def tunnel(self, wgport):
        self.wrapper.tunnel(wgport)

    def restart(self):
        self.wrapper.restart()

    @property
    def source_port(self) -> int:
        if self._cached_source_port is None:
            self._cached_source_port = find_free_tcp_port()
        return self._cached_source_port

    @property
    def listen_port(self) -> int:
        if self._cached_listen_port is None:
            self._cached_listen_port = find_free_udp_port()
        return self._cached_listen_port

    @property
    def peer_ips(self) -> List[str]:
        return self.wrapper.peer_ips

    @property
    def peer_scheme(self) -> str:
        try:
            parsed = urlparse(self.endpoint)
            return parsed.scheme
        except Exception:
            return ""

    @property
    def peer_port(self):
        if self.peer_scheme == "http":
            return 80
        return 443



from gettext import gettext as _
from gettext import ngettext
from typing import Tuple

import gi
from eduvpn_common.main import WrappedError

from eduvpn.connection import Validity
from eduvpn.i18n import extract_translation
from eduvpn.utils import run_in_glib_thread

GtkAvailable = True
try:
    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk  # type: ignore
except ValueError:
    GtkAvailable = False


IGNORE_ID = -13
QUIT_ID = -14


@run_in_glib_thread
def style_widget(widget, class_name: str, style: str):
    assert GtkAvailable
    style_context = widget.get_style_context()
    provider = Gtk.CssProvider.new()
    provider.load_from_data(f".{class_name} {{{style}}}".encode("utf-8"))  # type: ignore
    style_context.add_provider(provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
    style_context.add_class(class_name.split(":")[0])


def translated_error(error: Exception) -> str:
    translated = str(error)
    if isinstance(error, WrappedError):
        translated = extract_translation(error.translations)
    return translated


def should_show_error(error: Exception):
    if isinstance(error, WrappedError):
        return not error.misc
    return True


def get_validity_text(validity: Validity) -> Tuple[bool, str, str]:
    if validity is None:
        return (False, _("N/A"), _("Valid for: <b>unknown</b>"))
    if validity.is_expired:
        return (True, _("expired"), _("This session has expired"))
    delta = validity.remaining
    days = delta.days
    hours, remainder = divmod(delta.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    if days == 0:
        if hours == 0:
            if minutes == 0:
                return (
                    False,
                    f"{seconds}s",
                    ngettext(
                        "Valid for <b>{0} second</b>",
                        "Valid for <b>{0} seconds</b>",
                        seconds,
                    ).format(seconds),
                )
            else:
                # Round up minutes
                round_up_minutes = minutes
                label_text = "Valid for"
                if seconds > 30:
                    round_up_minutes += 1
                    label_text = "Valid for less than"
                if seconds < 30:
                    label_text = "Valid for more than"
                return (
                    False,
                    f"{minutes}m {seconds}s",
                    ngettext(
                        label_text + ": <b>{0} minute</b>",
                        label_text + ": <b>{0} minutes</b>",
                        round_up_minutes,
                    ).format(round_up_minutes),
                )
        else:
            # Round up hours
            round_up_hours = hours
            label_text = "Valid for"
            if minutes > 30:
                round_up_hours += 1
                label_text = "Valid for less than"
            if minutes < 30:
                label_text = "Valid for more than"
            return (
                False,
                f"{hours}h {minutes}m {seconds}s",
                ngettext(
                    label_text + ": <b>{0} hour</b>",
                    label_text + ": <b>{0} hours</b>",
                    round_up_hours,
                ).format(round_up_hours),
            )
    else:
        round_up_days = days
        label_text = "Valid for"
        if hours > 12:
            round_up_days += 1
            label_text = "Valid for less than"
        if minutes < 12:
            label_text = "Valid for more than"
        return (
            False,
            f"{days}d {hours}h {minutes}m {seconds}s",
            ngettext(
                label_text + ": <b>{0} day</b>",
                label_text + ": <b>{0} days</b>",
                round_up_days,
            ).format(round_up_days),
        )


@run_in_glib_thread
def show_ui_component(component, show: bool) -> None:
    """
    Set the visibility of a UI component.
    """
    if show:
        component.show()  # type: ignore
    else:
        component.hide()  # type: ignore


def link_markup(link: str) -> str:
    try:
        _scheme, rest = link.split(":", 1)
        if rest.startswith("//"):
            rest = rest[2:]
    except ValueError:
        return link
    else:
        return f'<a href="{link}">{rest}</a>'


@run_in_glib_thread
def show_error_dialog(parent, name: str, title: str, message: str, only_quit: bool = False):
    assert GtkAvailable
    dialog = Gtk.MessageDialog(  # type: ignore
        parent=parent,
        type=Gtk.MessageType.INFO,  # type: ignore
        title=name,
        message_format=title,
    )

    if not only_quit:
        dialog.add_buttons(_("Ignore and continue"), IGNORE_ID)  # type: ignore

    dialog.add_buttons(  # type: ignore
        _("Quit client"),
        QUIT_ID,
    )
    dialog.format_secondary_text(message)  # type: ignore
    dialog.show()  # type: ignore
    close = dialog.run()  # type: ignore
    dialog.destroy()  # type: ignore
    if close == QUIT_ID or only_quit:
        parent.close()



"""
This module contains code to maintain a simple metadata storage in ~/.config/eduvpn/
"""

from typing import Optional

from eduvpn.settings import CONFIG_DIR_MODE, CONFIG_PREFIX
from eduvpn.utils import get_logger

logger = get_logger(__name__)


def get_setting(variant, what: str) -> Optional[str]:
    p = (variant.config_prefix / what).expanduser()
    if p.exists():
        return open(p, "r").read().strip()
    else:
        return None


def is_config_dir_permissions_correct() -> bool:
    return CONFIG_PREFIX.stat().st_mode & 0o777 == CONFIG_DIR_MODE


def check_config_dir_permissions():
    if not is_config_dir_permissions_correct():
        logger.warning(
            f"The permissions for the config dir ({CONFIG_PREFIX}) are not as expected, it may be world readable!"
        )


def ensure_config_dir_exists():
    """
    Ensure the config directory exists with the correct permissions.
    """
    CONFIG_PREFIX.mkdir(parents=True, exist_ok=True, mode=CONFIG_DIR_MODE)
    check_config_dir_permissions()


def set_setting(variant, what: str, value: str):
    p = (variant.config_prefix / what).expanduser()
    ensure_config_dir_exists()
    with open(p, "w") as f:
        f.write(value)


def get_uuid(variant) -> Optional[str]:
    """
    Read the UUID of the last generated eduVPN Network Manager connection.
    """
    return get_setting(variant, "uuid")


def set_uuid(variant, uuid: str):
    """
    Write the eduVPN network manager connection UUID to disk.
    """
    set_setting(variant, "uuid", uuid)



import argparse

# readline is used! It is for going up and down in interactive mode
import readline  # noqa: F401
import signal
import sys
from functools import partial
from typing import Optional

import eduvpn_common.main as common
from eduvpn_common import __version__ as commonver
from eduvpn_common.state import State, StateType

import eduvpn.nm as nm
from eduvpn import __version__
from eduvpn.app import Application
from eduvpn.connection import parse_expiry
from eduvpn.i18n import retrieve_country_name
from eduvpn.server import (
    InstituteServer,
    Profile,
    SecureInternetServer,
    Server,
    ServerDatabase,
    ServerGroup,
    group_servers,
)
from eduvpn.settings import (
    CLIENT_ID,
    CONFIG_DIR_MODE,
    CONFIG_PREFIX,
    LETSCONNECT_CLIENT_ID,
    LETSCONNECT_CONFIG_PREFIX,
)
from eduvpn.ui.utils import get_validity_text, should_show_error, translated_error
from eduvpn.utils import (
    FAILOVERED_STATE,
    ONLINEDETECT_STATE,
    cmd_transition,
    init_logger,
    run_in_background_thread,
)
from eduvpn.variants import EDUVPN, LETS_CONNECT, ApplicationVariant


def get_grouped_index(servers, index, sort=True):
    if index < 0 or index >= len(servers):
        return None

    grouped_servers = group_servers(servers)
    if not sort:
        return (
            grouped_servers[ServerGroup.INSTITUTE_ACCESS]
            + grouped_servers[ServerGroup.SECURE_INTERNET]
            + grouped_servers[ServerGroup.OTHER]
        )[index]
    servers_added = sorted(grouped_servers[ServerGroup.INSTITUTE_ACCESS], key=lambda x: str(x))
    servers_added += sorted(grouped_servers[ServerGroup.SECURE_INTERNET], key=lambda x: str(x))
    servers_added += sorted(grouped_servers[ServerGroup.OTHER], key=lambda x: str(x))
    return servers_added[index]


def ask_profiles(setter, profiles, current: Optional[Profile] = None) -> bool:
    if len(profiles.profiles) == 1:
        _id, name = list(profiles.profiles.items())[0]
        print("There is only a single profile for this server:", name)
        setter(_id)
        return False

    # Multiple profiles, print the index
    sorted_profiles = sorted(profiles.profiles.items(), key=lambda pair: pair[1])
    # Multiple profiles, print the index
    index = 0
    choices = []
    for _id, profile in sorted_profiles:
        print(f"[{index + 1}]: {str(profile)}")
        choices.append(_id)
        index += 1

    # And ask for the 1 based index
    while True:
        profile_nr = input("Please select a profile number to continue connecting: ")
        try:
            profile_index = int(profile_nr)

            if profile_index < 1 or profile_index > len(choices):
                print(f"Invalid profile choice: {profile_index}")
                continue
            chosen = choices[profile_index - 1]
            if current is None or chosen != current.identifier:
                setter(choices[profile_index - 1])
                return True
            print("Selected profile is the same as the current profile")
            return False
        except ValueError:
            print(f"Input is not a number: {profile_nr}")


def ask_locations(setter, locations):
    # Create tuples of country name, location id
    location_tuples = []
    for loc in locations:
        country_name = retrieve_country_name(loc)
        location_tuples.append((country_name, loc))

    # Sort them and then print
    location_tuples.sort(key=lambda k: k[0])
    for index, location in enumerate(location_tuples):
        print(f"[{index + 1}]: {location[0]}")

    while True:
        location_nr = input("Please select a location to continue: ")
        try:
            location_index = int(location_nr)

            if location_index < 1 or location_index > len(location_tuples):
                print(f"Invalid location choice: {location_index}")
                continue
            setter(location_tuples[location_index - 1][1])
            return
        except ValueError:
            print(f"Input is not a number: {location_nr}")


class CommandLine:
    def __init__(self, name: str, variant: ApplicationVariant, common):
        self.name = name
        self.variant = variant
        self.common = common
        self.app = Application(variant, common)
        self.nm_manager = self.app.nm_manager
        self.server_db = ServerDatabase(common, variant.use_predefined_servers)
        self.transitions = CommandLineTransitions(self.app, self.nm_manager)
        self.skip_yes = False
        self.common.register_class_callbacks(self.transitions)

    def ask_yes(self, label) -> bool:
        if self.skip_yes:
            return True
        while True:
            yesno = input(label)

            if yesno in ["y", "yes"]:
                return True
            if yesno in ["n", "no"]:
                return False
            print(f'Input "{yesno}" is not valid')

    def ask_server_input(self, servers, fallback_search=False, query=""):
        print("Multiple servers found:")
        self.list_groups(group_servers(servers), query == "")

        while True:
            if fallback_search:
                server_nr = input("\nPlease select a server number or enter a search query: ")
            else:
                server_nr = input("\nPlease select a server number: ")
            try:
                server_index = int(server_nr)
                server = get_grouped_index(servers, server_index - 1, query == "")
                if not server:
                    print(f"Invalid server number: {server_index}")
                else:
                    return server
            except ValueError:
                print(f"Input is not a number: {server_nr}")

                if fallback_search:
                    print("Using the term as a search query instead...")
                    server = self.get_server_search(server_nr)
                    if server:
                        return server

    def get_server_search(self, search_query):
        servers = self.get_discovery()
        servers = list(self.server_db.search_predefined(search_query))
        if search_query.count(".") >= 2:
            servers.append(Server(search_query, search_query))

        if len(servers) == 0:
            print(f"No servers found with query: {search_query}")
            return None

        if len(servers) == 1:
            server = servers[0]
            print(f'One server found: "{str(server)}" ({server.category_id})')
            is_yes = self.ask_yes("Do you want to connect to it (y/n): ")

            if is_yes:
                return servers[0]
            else:
                return None

        return self.ask_server_input(servers, query=search_query)

    def connect_server(self, server, prefer_tcp: bool):
        if getattr(server, "delisted", False):
            print(
                f"Server is not longer available: {str(server)}. Please contact your IT department",
                file=sys.stderr,
            )
            return
        self.common.set_state(State.MAIN)

        def connect(callback=None):
            def connect_cb(success: bool = False):
                if callback:
                    callback()

            @run_in_background_thread("connect")
            def connect_background(server):
                try:
                    # Connect to the server and ensure it exists
                    not_exists = self.server_db.has(server) is None
                    if not_exists:
                        self.app.model.add(server)
                    self.app.model.connect(
                        server,
                        connect_cb,
                        prefer_tcp=prefer_tcp,
                    )
                except Exception as e:
                    if should_show_error(e):
                        print("Error connecting:", translated_error(e), file=sys.stderr)
                    if callback:
                        callback()

            connect_background(server)

        if not server:
            print("No server found to connect to, exiting...")
            return

        nm.action_with_mainloop(connect)
        if self.nm_manager.proxy is not None and self.common.in_state(State.CONNECTED):
            # make sure ctrl+c disconnects and then calls our original sighandler
            def quit_sigint(signal, frame):
                self.disconnect()
                self.app.cleanup_sigint(signal, frame)
                sys.exit(0)

            signal.signal(signal.SIGINT, quit_sigint)
            input(
                "\nYou are connected but we are proxying your connection over TCP, exiting the CLI will close the VPN. Press a key to exit...\n\n"
            )
            print("disconnecting and exiting...")
            self.disconnect()

    def ask_server_custom(self):
        custom = input("Enter a URL to connect to: ")
        return Server(custom, custom)

    def get_discovery(self):
        try:
            self.server_db.disco_update()
        except Exception as e:
            print(f"Failed to get discovery list: {str(e)}", file=sys.stderr)
            return []
        return self.server_db.disco

    def ask_server(self):
        if not self.variant.use_predefined_servers:
            return self.ask_server_custom()

        if self.server_db.configured:
            is_yes = self.ask_yes("Do you want to connect to an already configured server? (y/n): ")

            if is_yes:
                return self.ask_server_input(self.server_db.configured)

        servers = self.get_discovery()
        return self.ask_server_input(servers, fallback_search=True)

    def parse_server(self, variables):
        search_query = variables.get("search", None)
        url = variables.get("url", None)
        custom_url = variables.get("custom_url", None)
        org_id = variables.get("orgid", None)
        number = variables.get("number", None)
        number_all = variables.get("number_all", None)

        if search_query:
            server = self.get_server_search(search_query)
        elif url:
            server = InstituteServer(url, {"en": "Institute Server"}, [], None)
        elif org_id:
            server = SecureInternetServer(
                org_id,
                {"en": "Organisation Server"},
                [],
                None,
                "NL",
                [],
            )
        elif custom_url:
            server = Server(custom_url, {"en": "Custom Server"})
        elif number is not None:
            server = get_grouped_index(self.server_db.configured, number - 1)
            if not server:
                print(f"Configured server with number: {number} does not exist")
        elif number_all is not None:
            servers = self.get_discovery()
            server = get_grouped_index(servers, number_all - 1)
            if not server:
                print(
                    f"Server with number: {number_all} does not exist. Maybe the server list had an update? Make sure to re-run list --all"
                )
        return server

    def status(self, _args={}):
        if not self.common.in_state(State.CONNECTED):
            print("You are currently not connected to a server", file=sys.stderr)
            return False

        current = self.app.model.current_server
        print(f'Connected to: "{str(current)}" ({current.category_id})')
        validity = parse_expiry(self.common.get_expiry_times())
        valid_for = get_validity_text(validity)[1].replace("<b>", "").replace("</b>", "")
        print(f"Valid for: {valid_for}")
        print(f"Current profile: {str(current.profiles.current)}")
        if isinstance(current, SecureInternetServer):
            print(f"Current location: {retrieve_country_name(current.country_code)}")
        print(f"VPN Protocol: {self.nm_manager.protocol}")

    def connect(self, variables={}):
        if self.common.in_state(State.CONNECTED):
            print(
                "You are already connected to a server, please disconnect first",
                file=sys.stderr,
            )
            return False

        if not variables:
            server = self.ask_server()
        else:
            server = self.parse_server(variables)

        prefer_tcp = variables.get("tcp", False)

        return self.connect_server(server, prefer_tcp)

    def disconnect(self, _arg={}):
        if not self.common.in_state(State.CONNECTED):
            print("You are not connected to a server", file=sys.stderr)
            return False

        def disconnect(callback=None):
            try:
                self.app.model.deactivate_connection(callback)
            except Exception as e:
                if should_show_error(e):
                    print("An error occurred while trying to disconnect")
                    print("Error disconnecting:", translated_error(e), file=sys.stderr)
                if callback:
                    callback()

        nm.action_with_mainloop(disconnect)

    def list_groups(self, grouped_servers, sort=True):
        total_servers = 1
        if len(grouped_servers[ServerGroup.INSTITUTE_ACCESS]) > 0:
            print("============================")
            print("Institute Access Servers")
            print("============================")

        ias = grouped_servers[ServerGroup.INSTITUTE_ACCESS]
        if sort:
            ias = sorted(ias, key=lambda x: str(x))
        for institute in ias:
            prefix = ""
            if getattr(institute, "delisted", False):
                prefix += "No longer available: "
            print(f"[{total_servers}]: {prefix}{str(institute)}")
            total_servers += 1

        sis = grouped_servers[ServerGroup.SECURE_INTERNET]
        if sis:
            sis = sorted(sis, key=lambda x: str(x))
        if len(grouped_servers[ServerGroup.SECURE_INTERNET]) > 0:
            print("============================")
            print("Secure Internet Server")
            print("============================")
        for secure in sis:
            prefix = ""
            if getattr(secure, "delisted", False):
                prefix += "No longer available: "
            print(f"[{total_servers}]: {prefix}{str(secure)}")
            total_servers += 1

        custs = grouped_servers[ServerGroup.OTHER]
        if custs:
            custs = sorted(custs, key=lambda x: str(x))
        if len(grouped_servers[ServerGroup.OTHER]) > 0:
            print("============================")
            print("Custom Servers")
            print("============================")
        for custom in custs:
            print(f"[{total_servers}]: {str(custom)}")
            total_servers += 1
        if total_servers > 1:
            print("The number for the server is in [brackets]")

    def list(self, args={}):
        servers = self.server_db.configured
        if args.get("all"):
            servers = self.get_discovery()
        self.list_groups(group_servers(servers))

    def remove_server(self, server):
        if not server:
            print("No server chosen to remove", file=sys.stderr)
            return False

        self.app.model.remove(server)

    def change_profile(self, args={}):
        if not self.common.in_state(State.CONNECTED):
            print(
                "Please connect to a server first before changing profiles",
                file=sys.stderr,
            )
            return False

        server = self.app.model.current_server

        setter = self.app.model.set_profile
        print("Current profile:", server.profiles.current)

        if not ask_profiles(setter, server.profiles, server.profiles.current):
            return

        @run_in_background_thread("change-profile-reconnect")
        def reconnect(callback=None):
            try:
                self.app.model.reconnect(callback)
            except Exception as e:
                if should_show_error(e):
                    print("An error occurred while trying to reconnect for profile change")
                    print("Error reconnecting:", translated_error(e), file=sys.stderr)
                if callback:
                    callback()

        print("Reconnecting with the configured profile...")
        nm.action_with_mainloop(reconnect)

    def change_location(self, args={}):
        if not self.common.in_state(State.CONNECTED):
            print(
                "Please connect to a Secure Internet server first before changing locations",
                file=sys.stderr,
            )
            return False

        server = self.app.model.current_server
        if not isinstance(server, SecureInternetServer):
            print(
                "The currently connected server is not a Secure Internet server",
                file=sys.stderr,
            )
            return False

        print("Current location:", retrieve_country_name(server.country_code))

        def setter(loc):
            self.app.common.set_secure_location(server.org_id, loc)

        ask_locations(setter, server.locations)

        @run_in_background_thread("change-location-reconnect")
        def reconnect(callback=None):
            try:
                self.app.model.reconnect(callback)
            except Exception as e:
                if should_show_error(e):
                    print("An error occurred while trying to reconnect for location change")
                    print("Error reconnecting:", translated_error(e), file=sys.stderr)
                if callback:
                    callback()

        print("Reconnecting with the configured location...")
        nm.action_with_mainloop(reconnect)

    def renew(self, args={}):
        if not self.common.in_state(State.CONNECTED):
            print("Please connect to a server first before renewing", file=sys.stderr)
            return False

        def renew(callback):
            def renew_cb(success: bool = False):
                callback()

            @run_in_background_thread("renew")
            def renew_background():
                try:
                    self.app.model.renew_session(renew_cb)
                except Exception as e:
                    if should_show_error(e):
                        print("An error occurred while trying to renew")
                        print("Error renewing:", translated_error(e), file=sys.stderr)
                    if callback:
                        callback()

            renew_background()

        print("Disconnecting and renewing...")
        nm.action_with_mainloop(renew)

    def remove(self, args={}):
        if self.common.in_state(State.CONNECTED):
            print(
                "Please disconnect from your server before doing any changes",
                file=sys.stderr,
            )
            return False

        if not self.server_db.configured:
            print("There are no servers configured to remove", file=sys.stderr)
            return False

        if not args:
            server = self.ask_server_input(self.server_db.configured)
        else:
            number = args.get("number", None)
            if number is None:
                print("Please enter a number")
                return
            server = get_grouped_index(self.server_db.configured, number - 1)
            if not server:
                print(f"Configured server with number: {number} does not exist")
        return self.remove_server(server)

    def help_interactive(self, commands):
        command_keys = sorted(list(commands.keys()) + ["help"])
        print(f"Available commands: {', '.join(command_keys)}")

    def update_state(self, initial: bool = False):
        def update_state_callback(callback):
            state = self.nm_manager.connection_state
            self.app.on_network_update_callback(state, initial)

            # This exits the main loop and gives back control to the CLI
            callback()

        nm.action_with_mainloop(update_state_callback)

    def interactive(self, _):
        # Show a title and the help
        print(f"Welcome to the {self.name} interactive commandline")
        # Execute the right command
        commands = {
            "change-profile": self.change_profile,
            "connect": self.connect,
            "disconnect": self.disconnect,
            "renew": self.renew,
            "remove": self.remove,
            "status": self.status,
            "list": self.list,
            "quit": lambda: print("Exiting..."),
        }
        if self.variant.use_predefined_servers:
            commands["change-location"] = self.change_location
        self.help_interactive(commands)
        command = ""
        while command != "quit":
            # Ask for the command to execute
            try:
                command = input(f"[{self.name}]: ")
            except EOFError:
                command = "quit"

            # Update the state right before we execute
            self.update_state()

            func = commands.get(command, partial(self.help_interactive, commands))
            func()

    def start(self):
        parser = argparse.ArgumentParser(description=f"The {self.name} command line client")
        parser.set_defaults(func=lambda _: parser.print_usage())
        parser.add_argument("-d", "--debug", action="store_true", help="enable debugging")
        parser.add_argument("-y", "--yes", action="store_true", help="answer yes for y/n prompts")
        parser.add_argument("-v", "--version", action="store_true", help="get version info")
        subparsers = parser.add_subparsers(title="subcommands")

        interactive_parser = subparsers.add_parser("interactive", help="an interactive version of the command line")
        interactive_parser.set_defaults(func=self.interactive)

        renew_parser = subparsers.add_parser("renew", help="renew the validity for the currently connected server")
        renew_parser.set_defaults(func=self.renew)

        if self.variant.use_predefined_servers:
            change_profile_parser = subparsers.add_parser(
                "change-location",
                help="change the location for the currently connected secure internet server",
            )
            change_profile_parser.set_defaults(func=self.change_location)

        change_profile_parser = subparsers.add_parser(
            "change-profile",
            help="change the profile for the currently connected server",
        )
        change_profile_parser.set_defaults(func=self.change_profile)

        connect_parser = subparsers.add_parser("connect", help="connect to a server")
        connect_parser.add_argument(
            "-t",
            "--tcp",
            action="store_true",
            help="prefer to connect using TCP if available. Useful if your network blocks UDP connections and the client/NetworkManager does not properly detect issues",
        )
        connect_group = connect_parser.add_mutually_exclusive_group(required=True)
        if self.variant.use_predefined_servers:
            connect_group.add_argument(
                "-s",
                "--search",
                type=str,
                help="connect to a new server by searching for one",
            )
            connect_group.add_argument(
                "-o",
                "--orgid",
                type=str,
                help="connect to a new secure internet server using the organisation ID",
            )
            connect_group.add_argument(
                "-u",
                "--url",
                type=str,
                help="connect to a new institute access server using the URL",
            )
        connect_group.add_argument(
            "-c",
            "--custom-url",
            type=str,
            help="connect to a new custom server using the URL",
        )
        connect_group.add_argument(
            "-n",
            "--number",
            type=int,
            help="connect to an already configured server using the number. Run the 'list' subcommand to see the currently configured servers with their number",
        )
        if self.variant.use_predefined_servers:
            connect_group.add_argument(
                "-a",
                "--number-all",
                type=int,
                help="connect to a new server using the number for all servers. Run the 'list --all' command to see all the available servers with their number",
            )
        connect_group.set_defaults(func=lambda args: self.connect(vars(args)))

        disconnect_parser = subparsers.add_parser("disconnect", help="disconnect the currently active server")
        disconnect_parser.set_defaults(func=self.disconnect)

        list_parser = subparsers.add_parser("list", help="list all configured servers")
        if self.variant.use_predefined_servers:
            list_parser.add_argument("--all", action="store_true", help="list all available servers")
        list_parser.set_defaults(func=lambda args: self.list(vars(args)))

        remove_parser = subparsers.add_parser("remove", help="remove a configured server")
        remove_parser.add_argument(
            "-n",
            "--number",
            type=int,
            required=True,
            help="remove a configured server by number",
        )
        remove_parser.set_defaults(func=lambda args: self.remove(vars(args)))

        status_parser = subparsers.add_parser("status", help="see the current status of eduVPN")
        status_parser.set_defaults(func=self.status)

        parsed = parser.parse_args()

        if parsed.version:
            print(f"eduVPN CLI version: {__version__} with eduvpn-common version: {commonver}")
            return

        init_logger(parsed.debug, self.variant.logfile, CONFIG_DIR_MODE)

        # Skip yes/no prompts if given
        self.skip_yes = parsed.yes

        # Register the common library
        self.app.model.register()

        # Update the state by asking NetworkManager
        self.update_state(True)

        # Run the command
        parsed.func(parsed)

        # Deregister the library
        self.common.deregister()


class CommandLineTransitions:
    def __init__(self, app, nm_manager):
        self.app = app
        self.nm_manager = nm_manager

    @cmd_transition(State.ASK_LOCATION, StateType.ENTER)
    def on_ask_location(self, old_state: State, data):
        print("This Secure Internet Server has multiple available locations. Please choose one to continue...")
        setter, locations = data
        ask_locations(setter, locations)

    @cmd_transition(State.ASK_PROFILE, StateType.ENTER)
    def on_ask_profile(self, old_state: State, data):
        print("This server has multiple profiles. Please choose one to continue...")
        setter, profiles = data
        ask_profiles(setter, profiles)

    @cmd_transition(State.OAUTH_STARTED, StateType.ENTER)
    def on_oauth_started(self, old_state: State, url: str):
        print(f"Authorization needed. Your browser has been opened with url: {url}")

    @cmd_transition(ONLINEDETECT_STATE, StateType.ENTER)  # type: ignore[arg-type]
    def on_online_detection(self, old_state: State, data: str):
        print("Connected, but we are testing your VPN connection...")

    @cmd_transition(FAILOVERED_STATE, StateType.ENTER)  # type: ignore[arg-type]
    def on_failovered(self, old_state: State, data: str):
        print("The connection has switched to a new VPN protocol...")


def eduvpn():
    _common = common.EduVPN(CLIENT_ID, __version__, str(CONFIG_PREFIX))
    cmd = CommandLine("eduVPN", EDUVPN, _common)
    cmd.start()


def letsconnect():
    _common = common.EduVPN(LETSCONNECT_CLIENT_ID, __version__, str(LETSCONNECT_CONFIG_PREFIX))
    cmd = CommandLine("Let's Connect!", LETS_CONNECT, _common)
    cmd.start()



import enum
import json
import logging
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union, cast

from eduvpn_common.main import ServerType

from eduvpn.discovery import (
    DiscoOrganization,
    DiscoServer,
    parse_disco_organizations,
    parse_disco_servers,
)
from eduvpn.i18n import extract_translation
from eduvpn.settings import IMAGE_PREFIX

logger = logging.getLogger(__name__)


class Profile:
    """The class that represents a server profile.
    :param: identifier: str: The identifier (id) of the profile
    :param: display_name: str: The display name of the profile
    :param: default_gateway: str: Whether or not this profile should have the default gateway set
    :param: priority: int: The priority of the profile for sorting in the UI
    """

    def __init__(self, identifier: str, display_name: Dict[str, str], default_gateway: bool, priority: int):
        self.identifier = identifier
        self.display_name = display_name
        self.default_gateway = default_gateway
        self.priority = priority

    def __str__(self):
        return extract_translation(self.display_name)

    def __lt__(self, o) -> bool:
        if self.priority == o.priority:
            return str(self) < str(o)
        return self.priority > o.priority


class Profiles:
    """The class that represents a list of profiles
    :param: profiles: List[Profile]: A list of profiles
    :param: current: int: The current profile index
    """

    def __init__(self, profiles: Dict[str, Profile], current: str):
        self.profiles = profiles
        self.current_id = current

    @property
    def current(self) -> Optional[Profile]:
        """Get the current profile if there is any
        :return: The profile if there is a current one (meaning the index is valid)
        :rtype: Optional[Profile]
        """
        if self.current_id not in self.profiles:
            return None
        return self.profiles[self.current_id]


class Server:
    """The class that represents a server. Use this for a custom server
    :param: url: str: The base URL of the server. In case of secure internet (supertype) this is the organisation ID URL
    :param: display_name: str: The display name of the server
    :param: profiles: Optional[Profiles]: The profiles if there are any already obtained, defaults to None
    """

    def __init__(
        self,
        url: str,
        display_name: Dict[str, str],
        profiles: Optional[Profiles] = None,
        delisted: bool = False,
    ):
        self.url = url
        self.display_name = display_name
        self.profiles = profiles
        self.delisted = delisted

    def __str__(self) -> str:
        return extract_translation(self.display_name)

    @property
    def identifier(self) -> str:
        return self.url

    @property
    def category_id(self) -> ServerType:
        return ServerType.CUSTOM


class InstituteServer(Server):
    """The class that represents an Institute Access Server
    :param: url: str: The base URL of the Institute Access Server
    :param: display_name: str: The display name of the Institute Access Server
    :param: support_contact: List[str]: The list of support contacts
    :param: profiles: Profiles: The profiles of the server
    """

    def __init__(
        self,
        url: str,
        display_name: Dict[str, str],
        support_contact: List[str],
        profiles: Profiles,
        delisted: bool = False,
    ):
        super().__init__(url, display_name, profiles, delisted)
        self.support_contact = support_contact

    @property
    def category_id(self) -> ServerType:
        return ServerType.INSTITUTE_ACCESS


class SecureInternetServer(Server):
    """The class that represents a Secure Internet Server
    :param: org_id: str: The organization ID of the Secure Internet Server as returned by Discovery
    :param: display_name: str: The display name of the server
    :param: support_contact: List[str]: The list of support contacts of the server
    :param: locations: List[str]: The list of secure internet locations
    :param: profiles: Profiles: The list of profiles that the server has
    :param: country_code: str: The country code of the server
    :param: locations: List[str]: The list of secure internet locations
    """

    def __init__(
        self,
        org_id: str,
        display_name: Dict[str, str],
        support_contact: List[str],
        profiles: Profiles,
        country_code: str,
        locations: List[str],
        delisted: bool = False,
    ):
        super().__init__(org_id, display_name, profiles, delisted)
        self.org_id = org_id
        self.support_contact = support_contact
        self.country_code = country_code
        self.locations = locations

    @property
    def category_id(self) -> ServerType:
        return ServerType.SECURE_INTERNET


def parse_secure_internet(si: dict) -> Optional[SecureInternetServer]:
    profiles = parse_profiles(si["profiles"])
    locations = si.get("locations", [])
    return SecureInternetServer(
        si["identifier"],
        si["display_name"],
        si.get("support_contacts", []),
        profiles,
        si["country_code"],
        locations,
        si["delisted"],
    )


def parse_current_server(server_json: str) -> Optional[Server]:
    d = json.loads(server_json)
    t = ServerType(d["server_type"])
    if t == ServerType.UNKNOWN:
        return None
    if t == ServerType.INSTITUTE_ACCESS:
        i = d["institute_access_server"]
        profiles = parse_profiles(i["profiles"])
        return InstituteServer(i["identifier"], i["display_name"], i["support_contacts"], profiles)
    if t == ServerType.SECURE_INTERNET:
        si = d["secure_internet_server"]
        return parse_secure_internet(si)
    if t == ServerType.CUSTOM:
        c = d["custom_server"]
        profiles = parse_profiles(c["profiles"])
        return Server(c["identifier"], c["display_name"], profiles)


def parse_profiles(profiles: dict) -> Profiles:
    returned = {}
    profile_map = profiles.get("map", {})
    for k, v in profile_map.items():
        # TODO: Default gateway
        returned[k] = Profile(k, v["display_name"], False, int(v.get("priority", 0)))
    return Profiles(returned, profiles["current"])


def parse_required_transition(transition_json: str, get: Optional[Callable] = None) -> Tuple[int, Any]:
    transition = json.loads(transition_json)
    data_parsed = transition["data"]
    if get is not None:
        data_parsed = get(data_parsed)
    cookie = transition["cookie"]
    return cookie, data_parsed


def parse_servers(server_json: str) -> List[Server]:
    d = json.loads(server_json)

    institutes = d.get("institute_access_servers", [])
    servers: List[Server] = []
    for i in institutes:
        profiles = parse_profiles(i["profiles"])
        servers.append(
            InstituteServer(
                i["identifier"],
                i["display_name"],
                i.get("support_contacts", []),
                profiles,
                i["delisted"],
            )
        )

    customs = d.get("custom_servers", [])
    for i in customs:
        profiles = parse_profiles(i["profiles"])
        servers.append(Server(i["identifier"], i["display_name"], profiles))

    si = d.get("secure_internet_server", None)
    if si is not None:
        # right now we only support one secure internet server
        si_parsed = parse_secure_internet(si)
        if si_parsed is not None:
            servers.append(si_parsed)
    return servers


class StatusImage(enum.Enum):
    # The value is the image filename.
    DEFAULT = "desktop-default.png"
    CONNECTING = "desktop-connecting.png"
    CONNECTED = "desktop-connected.png"
    NOT_CONNECTED = "desktop-not-connected.png"

    @property
    def path(self) -> str:
        return IMAGE_PREFIX + self.value


class ServerDatabase:
    def __init__(self, wrapper, enable_discovery=True) -> None:
        self.wrapper = wrapper
        self.enable_discovery = enable_discovery
        self.cached: List[Union[DiscoServer, DiscoOrganization]] = []

    @property
    def disco(self):
        if not self.enable_discovery:
            return []
        return self.cached

    def disco_update(self, search="", cache: bool = False):
        if not self.enable_discovery:
            return
        disco_orgs = []
        if self.secure_internet is None:
            disco_orgs = parse_disco_organizations(self.wrapper.get_disco_organizations(cache, search))
        disco_servers = parse_disco_servers(self.wrapper.get_disco_servers(cache, search))
        ret_servers = cast(List[Union[DiscoServer, DiscoOrganization]], disco_orgs)
        ret_servers.extend(disco_servers)
        if search == "":
            self.cached = ret_servers
        return ret_servers

    def has(self, server) -> Optional[Server]:
        # The url attribute is always used as an identifier
        for s in self.configured:
            if server.identifier == s.identifier:
                return s
        return None

    @property
    def secure_internet(self) -> Optional[SecureInternetServer]:
        for server in self.configured:
            if isinstance(server, SecureInternetServer):
                return server
        return None

    @property
    def current(self):
        try:
            return parse_current_server(self.wrapper.get_current_server())
        except Exception as e:
            logger.debug(f"failed to get current server: {str(e)}")
            return None

    @property
    def configured(self):
        return parse_servers(self.wrapper.get_servers())

    def all(self):
        "Return all servers."
        return self.cached

    def search_predefined(self, query: str):
        "Return all servers that match the search query."
        return self.disco_update(query)

    def search_custom(self, query: str) -> Iterable[Server]:
        yield Server(query, query)  # type: ignore[arg-type]


class ServerGroup(enum.Enum):
    INSTITUTE_ACCESS = enum.auto()
    SECURE_INTERNET = enum.auto()
    OTHER = enum.auto()


def group_servers(servers):
    """
    Separate the servers into three groups.
    """
    groups: Dict[ServerGroup, List[Server]] = {  # type: ignore
        ServerGroup.INSTITUTE_ACCESS: [],
        ServerGroup.SECURE_INTERNET: [],
        ServerGroup.OTHER: [],
    }
    for server in servers:
        if isinstance(server, InstituteServer) or (
            isinstance(server, DiscoServer) and server.server_type == "institute_access"
        ):
            groups[ServerGroup.INSTITUTE_ACCESS].append(server)
        elif isinstance(server, SecureInternetServer) or isinstance(server, DiscoOrganization):
            groups[ServerGroup.SECURE_INTERNET].append(server)
        elif isinstance(server, Server):
            groups[ServerGroup.OTHER].append(server)
        else:
            continue
    return groups



import enum
import logging
import os
import time
import uuid
from configparser import ConfigParser
from contextlib import closing
from ipaddress import ip_address, ip_interface, ip_network
from pathlib import Path
from shutil import rmtree
from socket import AF_INET, AF_INET6, IPPROTO_TCP, SOCK_DGRAM, socket
from tempfile import mkdtemp
from typing import Any, Callable, Optional, TextIO

from eduvpn_common.main import EduVPN, Jar
from gi.repository.Gio import Cancellable, Task  # type: ignore

from eduvpn.storage import get_uuid, set_uuid
from eduvpn.utils import run_in_background_thread, run_in_glib_thread
from eduvpn.variants import ApplicationVariant

_logger = logging.getLogger(__name__)

LINUX_NET_FOLDER = Path("/sys/class/net")

try:
    import gi

    gi.require_version("NM", "1.0")
    from gi.repository import NM, GLib  # type: ignore
except (ImportError, ValueError):
    _logger.warning("Network Manager not available")
    NM = None


class ConnectionState(enum.Enum):
    CONNECTING = enum.auto()
    CONNECTED = enum.auto()
    DISCONNECTED = enum.auto()
    FAILED = enum.auto()
    UNKNOWN = enum.auto()

    @classmethod
    def from_vpn_state(cls, state: "NM.VpnConnectionState") -> "ConnectionState":
        if state in [
            NM.VpnConnectionState.CONNECT,
            NM.VpnConnectionState.IP_CONFIG_GET,
            NM.VpnConnectionState.PREPARE,
        ]:
            return cls.CONNECTING
        elif state is NM.VpnConnectionState.ACTIVATED:
            return cls.CONNECTED
        elif state is NM.VpnConnectionState.DISCONNECTED:
            return cls.DISCONNECTED
        elif state is NM.VpnConnectionState.FAILED:
            return cls.FAILED
        elif state is NM.VpnConnectionState.NEED_AUTH:
            return cls.FAILED
        elif state is NM.VpnConnectionState.UNKNOWN:
            return cls.UNKNOWN
        else:
            raise ValueError(state)

    @classmethod
    def from_active_state(cls, state: "NM.ActiveConnectionState") -> "ConnectionState":
        if state is NM.ActiveConnectionState.ACTIVATING:
            return cls.CONNECTING
        elif state is NM.ActiveConnectionState.ACTIVATED:
            return cls.CONNECTED
        elif state in [
            NM.ActiveConnectionState.DEACTIVATED,
            NM.ActiveConnectionState.DEACTIVATING,
        ]:
            return cls.DISCONNECTED
        elif state is NM.ActiveConnectionState.UNKNOWN:
            return cls.UNKNOWN
        else:
            raise ValueError(state)


def find_free_udp_port_exclude(exclude_port):
    max_retries = 10
    retries = 0
    while retries < max_retries:
        with closing(socket(AF_INET, SOCK_DGRAM)) as s:
            s.bind(("", 0))
            port = s.getsockname()[1]
            if port != exclude_port:
                return port
        retries += 1
    raise RuntimeError(f"Failed to find an available ProxyGuard port after {max_retries} retries")


# A manager for a manager :-)
class NMManager:
    def __init__(self, variant: ApplicationVariant, common_lib: Optional[EduVPN]):
        self.variant = variant
        self.proxy = None
        try:
            self._client = NM.Client.new(None)
            self.wg_gateway_ip: Optional[str] = None
        except Exception:
            self._client = None
        self.cancel_jar = Jar(lambda x: x.cancel())
        self.common_lib = common_lib

    @property
    def client(self) -> "NM.Client":
        if self._client is None:
            raise Exception("no client available")
        return self._client

    @property
    def available(self) -> bool:
        return self._client is not None

    @run_in_background_thread("proxyguard-tunnel")
    def proxy_tunnel(self, wg_listen_port):
        assert self.proxy is not None
        _logger.debug(f"tunneling proxyguard with WG listen port: {wg_listen_port}")
        try:
            self.proxy.tunnel(wg_listen_port)
        except Exception as e:
            self.proxy.forward_exception(e)

    # TODO: Move this somewhere else?
    def open_stats_file(self, filename: str) -> Optional[TextIO]:
        """
        Helper function to open a statistics network file
        """
        if not self.iface:
            return None
        filepath = LINUX_NET_FOLDER / self.iface / "statistics" / filename  # type: ignore
        if not filepath.is_file():
            return None
        return open(filepath, "r")

    # TODO: Move this somewhere else?
    def get_stats_bytes(self, filehandler: Optional[TextIO]) -> Optional[int]:
        """
        Helper function to get a statistics file to calculate the total data transfer
        """
        # If the interface is not set
        # or the file is not present, we cannot get the stat
        if not self.iface:
            # Warning was already shown
            return None
        if not filehandler:
            # Warning was already shown
            return None

        # Get the statistic from the file
        # and go to the beginning
        try:
            stat = int(filehandler.readline())
        except ValueError:
            stat = 0
        filehandler.seek(0)
        return stat

    @property
    def managed(self) -> bool:
        """
        check if any device of the primary connection is managed by Network Manager
        """
        active_connection = self.client.get_primary_connection()
        if active_connection is None:
            return False

        master_devices = active_connection.get_devices()
        if master_devices is None:
            return False

        return any(d.get_managed() for d in master_devices)

    @property
    def mtu(self) -> Optional[int]:
        protocol = self.protocol
        # For WireGuard, we can get the MTU from the device
        if protocol == "WireGuard":
            device = self.wireguard_device
            if device is None:
                return None
            return device.get_mtu()
        elif protocol == "OpenVPN":
            # TODO: How to query networkmanager for this?
            return 1500
        return None

    @property
    def uuid(self):
        return get_uuid(self.variant)

    @uuid.setter
    def uuid(self, new_uuid):
        set_uuid(self.variant, new_uuid)

    @property
    def connection_state(self) -> ConnectionState:
        connections = [
            connection for connection in self.client.get_active_connections() if connection.get_uuid() == self.uuid
        ]
        if len(connections) == 1:
            connection = connections[0]
        else:
            return ConnectionState.DISCONNECTED
        if isinstance(connection, NM.VpnConnection):
            return ConnectionState.from_vpn_state(connection.get_vpn_state())
        elif isinstance(connection, NM.ActiveConnection):
            return ConnectionState.from_active_state(connection.get_state())
        else:
            _logger.warning(f"connection of unknown type: {connection!r}")
            return ConnectionState.UNKNOWN

    @property
    def active_connection(self) -> Optional["NM.ActiveConnection"]:
        """
        Gets the active connection for the current uuid
        """
        for connection in self.client.get_active_connections():
            if connection.get_uuid() == self.uuid:
                return connection
        return None

    @property
    def protocol(self) -> Optional[str]:
        """
        Get the VPN protocol as a string for the active connection
        """
        connection = self.active_connection
        if connection is None:
            return None
        type = connection.get_connection_type()
        if type == "vpn":
            return "OpenVPN"
        elif type == "wireguard":
            if self.proxy:
                return "WireGuard (ProxyGuard)"
            return "WireGuard"
        return None

    @property
    def iface(self) -> Optional[str]:
        """
        Get the interface as a string for an openvpn or wireguard connection if there is one
        """
        active_con = self.active_connection
        if not active_con:
            return None

        devices = active_con.get_devices()
        if not devices:
            return None

        # Not always a master device is configured
        # So get the interface for the first device we have
        return devices[0].get_iface()

    @property
    def ipv4_config(self) -> Optional["NM.IPConfig"]:
        """
        Get the ipv4 config for the active VPN connection
        """
        active_con = self.active_connection
        if not active_con:
            return None

        return active_con.get_ip4_config()

    @property
    def ipv4(self) -> Optional[str]:
        """
        Get the ipv4 address for an openvpn or wireguard connection as a string if there is one
        """
        ip4_config = self.ipv4_config
        if not ip4_config:
            return None

        addresses = ip4_config.get_addresses()
        if not addresses:
            return None

        return addresses[0].get_address()

    @property
    def ipv6(self) -> Optional[str]:
        """
        Get the ipv6 address for an openvpn or wireguard connection as a string if there is one
        """
        active_con = self.active_connection

        if not active_con:
            return None

        ip6_config = active_con.get_ip6_config()

        if not ip6_config:
            return None

        addresses = ip6_config.get_addresses()
        if not addresses:
            return None

        for addr in addresses:
            try:
                ipv6_str = addr.get_address()
                ipv6 = ip_address(ipv6_str)
                if not ipv6.is_link_local:
                    return ipv6_str
            except ValueError:
                continue
        return None

    @property
    def failover_endpoint_ip(self) -> Optional[str]:
        protocol = self.protocol
        if protocol == "OpenVPN":
            # if OpenVPN return the gateway from the ip config
            ipv4_config = self.ipv4_config
            if not ipv4_config:
                _logger.debug("no ipv4_config found in OpenVPN failover endpoint")
                return None
            return ipv4_config.get_gateway()
        elif protocol == "WireGuard":
            # if WireGuard return the cached IP
            if not self.wg_gateway_ip:
                _logger.debug("no wg gateway ip found in failover endpoint")
                return None
            return self.wg_gateway_ip
        else:
            _logger.debug(f"Unknown protocol: {protocol}")
            return None

    @property
    def existing_connection(self) -> Optional[str]:
        if not self.uuid:
            return None
        connection = self.client.get_connection_by_uuid(self.uuid)
        if connection is None:
            return None
        else:
            return self.uuid

    def ovpn_import(self, target: Path) -> Optional["NM.Connection"]:
        """
        Use the Network Manager VPN config importer to import an OpenVPN configuration file.
        """
        vpn_infos = [i for i in NM.VpnPluginInfo.list_load() if i.get_name() == "openvpn"]

        if len(vpn_infos) != 1:
            raise Exception(f"Expected one openvpn VPN plugins, got: {len(vpn_infos)}")

        conn = vpn_infos[0].load_editor_plugin().import_(str(target))
        conn.normalize()
        return conn

    def import_ovpn(self, ovpn: str) -> "NM.SimpleConnection":
        """
        Import the OVPN string into Network Manager.
        """
        target_parent = Path(mkdtemp())
        target = target_parent / f"{self.variant.name}.ovpn"
        _logger.debug(f"Writing configuration to {target}")
        with open(target, mode="w+t") as f:
            f.write(ovpn)
        connection = self.ovpn_import(target)
        rmtree(target_parent)
        return connection

    @run_in_glib_thread
    def add_connection(
        self,
        connection: "NM.Connection",
        callback: Optional[Callable] = None,
    ) -> None:
        _logger.debug("Adding new connection")
        c = self.new_cancellable()
        self.client.add_connection_async(
            connection=connection,
            save_to_disk=True,
            callback=add_connection_callback,
            cancellable=c,
            user_data=(self, c, callback),
        )

    def set_connection(
        self,
        new_connection: "NM.SimpleConnection",
        callback: Callable,
    ):
        new_connection = self.set_setting_ensure_permissions(new_connection)
        if self.existing_connection:

            def deleted(success: bool):
                if success:
                    self.add_connection(new_connection, callback)
                else:
                    callback(False)

            self.delete_connection(deleted)
        else:
            self.add_connection(new_connection, callback)

    def set_setting_ensure_permissions(self, con: "NM.SimpleConnection") -> "NM.SimpleConnection":
        s_con = con.get_setting_connection()
        s_con.add_permission("user", GLib.get_user_name(), None)
        con.add_setting(s_con)
        return con

    def start_openvpn_connection(self, ovpn: str, default_gateway, dns_search_domains, *, callback=None) -> None:
        _logger.debug("writing ovpn configuration to Network Manager")
        new_con = self.import_ovpn(ovpn)
        s_ip4 = new_con.get_setting_ip4_config()
        s_ip6 = new_con.get_setting_ip6_config()

        # avoid DNS leaks in default gateway
        # see man nm-settings for dns-priority
        # and https://systemd.io/RESOLVED-VPNS/
        if default_gateway:
            s_ip4.set_property(NM.SETTING_IP_CONFIG_DNS_PRIORITY, -2147483648)
            s_ip6.set_property(NM.SETTING_IP_CONFIG_DNS_PRIORITY, -2147483648)
            s_ip4.add_dns_search("~.")
            s_ip6.add_dns_search("~.")
        for i in dns_search_domains:
            s_ip4.add_dns_search(i)
            s_ip6.add_dns_search(i)
        s_ip4.set_property("never-default", not default_gateway)
        s_ip6.set_property("never-default", not default_gateway)
        new_con.add_setting(s_ip4)
        new_con.add_setting(s_ip6)

        self.proxy = None
        self.set_connection(new_con, callback)  # type: ignore

    def get_priorities(self, has_proxy: bool, has_lan: bool):
        # rule, proxy, lan, exclude
        if has_proxy:
            if has_lan:
                return [4, 3, 2, 1]
            else:
                return [2, 1, -1, -1]
        else:
            if has_lan:
                return [3, -1, 2, 1]
            else:
                return [1, -1, -1, -1]

    def start_wireguard_connection(  # noqa: C901
        self,
        config: ConfigParser,
        default_gateway,
        *,
        allow_wg_lan=False,
        proxy=None,
        callback=None,
    ) -> None:
        _logger.debug("writing wireguard configuration to Network Manager")
        ipv4s = []
        ipv6s = []
        lan_4 = []
        lan_6 = []
        # set LANs to whitelist
        # we cannot just add a ip rule that takes everything from main table except the default route
        # as we have Tunnelcrack/Tunnelvision attacks to worry about
        # from RFC1918: https://datatracker.ietf.org/doc/html/rfc1918#section-3
        if allow_wg_lan:
            lan_4 = [
                ip_network("10.0.0.0/8"),
                ip_network("172.16.0.0/12"),
                ip_network("192.168.0.0/16"),
            ]
            # from RFC4193 https://datatracker.ietf.org/doc/html/rfc4193#section-3.1
            lan_6 = [ip_network("fc00::/7")]
        # we check if traffic to the VPN ip is in one of the LAN ranges
        # if so, then we need to ensure that this traffic is going to the WG routing table
        # We can't just add a routing rule that it is always going to the WG routing table
        # as there could be public IPs here, and then we would get a routing loop
        needs_exclude_v4 = []
        needs_exclude_v6 = []

        self.wg_gateway_ip = None
        for ip in config["Interface"]["Address"].split(","):
            addr = ip_interface(ip.strip())
            if addr.version == 4:
                if not self.wg_gateway_ip:
                    net_str = str(addr.network)
                    if self.common_lib is None:
                        raise RuntimeError("common lib not available")
                    self.wg_gateway_ip = self.common_lib.calculate_gateway(net_str)
                ipv4s.append(NM.IPAddress(AF_INET, str(addr.ip), addr.network.prefixlen))

                # if this VPN v4 IP is in one of the LAN v4 ranges
                # we need to exclude this from the LAN v4 allow list
                for lan in lan_4:
                    if addr in lan:
                        needs_exclude_v4.append((str(addr.ip), addr.network.prefixlen))
            elif addr.version == 6:
                ipv6s.append(NM.IPAddress(AF_INET6, str(addr.ip), addr.network.prefixlen))
                # if this VPN v6 IP is in one of the LAN v6 ranges
                # we need to exclude this from the LAN v6 allow list
                for lan in lan_6:
                    if addr in lan:
                        needs_exclude_v6.append((str(addr.ip), addr.network.prefixlen))

        dns4 = []
        dns6 = []
        dns_hostnames = []

        # DNS entries are not required
        dns_entries = config["Interface"].get("DNS")
        if dns_entries:
            for dns_entry in dns_entries.split(","):
                stripped_entry = dns_entry.strip()
                try:
                    address = ip_address(stripped_entry)
                # The entry is not an ip but a hostname
                # They need to be added to dns search domains
                except ValueError:
                    dns_hostnames.append(stripped_entry)
                else:
                    if address.version == 4:
                        dns4.append(str(address))
                    elif address.version == 6:
                        dns6.append(str(address))

        profile = NM.SimpleConnection.new()
        s_con = NM.SettingConnection.new()
        s_con.set_property(NM.DEVICE_AUTOCONNECT, False)
        s_con.set_property(NM.SETTING_CONNECTION_ID, self.variant.name)
        s_con.set_property(NM.SETTING_CONNECTION_TYPE, "wireguard")
        s_con.set_property(NM.SETTING_CONNECTION_UUID, str(uuid.uuid4()))
        s_con.set_property(NM.SETTING_CONNECTION_INTERFACE_NAME, self.variant.translation_domain)

        # https://lazka.github.io/pgi-docs/NM-1.0/classes/WireGuardPeer.html#NM.WireGuardPeer
        peer = NM.WireGuardPeer.new()
        wg_endpoint = config["Peer"]["Endpoint"]

        if proxy is not None:
            if wg_endpoint != "":
                _logger.warning(f"ignoring WireGuard endpoint: {wg_endpoint} as we have to use ProxyGuard")
            wg_endpoint = f"127.0.0.1:{proxy.listen_port}"

        peer.set_endpoint(wg_endpoint, allow_invalid=False)

        psk = config["Peer"].get("PresharedKey")
        if psk:
            peer.set_preshared_key(psk, accept_invalid=False)
            # for some reason this needs to be set
            # AGENT_OWNED doesn't work
            peer.set_preshared_key_flags(NM.SettingSecretFlags.NONE)

        peer.set_public_key(config["Peer"]["PublicKey"], accept_invalid=False)
        for ip in config["Peer"]["AllowedIPs"].split(","):
            peer.append_allowed_ip(ip.strip(), accept_invalid=False)

        try:
            keepalivestr = config["Peer"].get("PersistentKeepalive", "0")
            keepalive = int(keepalivestr)
            if keepalive > 0:
                _logger.debug(f"setting persistent keepalive: {keepalive}")
                peer.set_persistent_keepalive(keepalive)
        except ValueError:
            _logger.warning(f"failed to parse persistent keepalive as integer: {keepalivestr}")

        s_ip4 = NM.SettingIP4Config.new()
        s_ip6 = NM.SettingIP6Config.new()

        for i in dns4:
            s_ip4.add_dns(i)
        for i in dns6:
            s_ip6.add_dns(i)

        # avoid DNS leaks in default gateway
        # see man nm-settings for dns-priority
        # and https://systemd.io/RESOLVED-VPNS/
        if default_gateway:
            s_ip4.set_property(NM.SETTING_IP_CONFIG_DNS_PRIORITY, -2147483648)
            s_ip6.set_property(NM.SETTING_IP_CONFIG_DNS_PRIORITY, -2147483648)
            s_ip4.add_dns_search("~.")
            s_ip6.add_dns_search("~.")
        for i in dns_hostnames:
            s_ip4.add_dns_search(i)
            s_ip6.add_dns_search(i)

        s_ip4.set_property("never-default", not default_gateway)
        s_ip6.set_property("never-default", not default_gateway)

        s_ip4.set_property(NM.SETTING_IP_CONFIG_METHOD, "manual")
        s_ip6.set_property(NM.SETTING_IP_CONFIG_METHOD, "manual")

        for i in ipv4s:
            s_ip4.add_address(i)
        for i in ipv6s:
            s_ip6.add_address(i)

        # https://lazka.github.io/pgi-docs/NM-1.0/classes/SettingWireGuard.html
        w_con = NM.SettingWireGuard.new()

        # manual routing
        w_con.set_property(NM.SETTING_WIREGUARD_IP4_AUTO_DEFAULT_ROUTE, 0)
        w_con.set_property(NM.SETTING_WIREGUARD_IP6_AUTO_DEFAULT_ROUTE, 0)

        # we set some sensible default with an override using EDUVPN_WG_FWMARK
        fwmark = int(os.environ.get("EDUVPN_WG_FWMARK", 51860))
        wg_listen_port = int(os.environ.get("EDUVPN_WG_LISTEN_PORT", 0))

        # if we are using proxyguard we need to know the port beforehand
        # just get an available port
        # Yes, this can race but there is not other option right now
        # the NM API just gives port 0 from device.get_listen_port()
        if proxy is not None and wg_listen_port == 0:
            wg_listen_port = find_free_udp_port_exclude(proxy.listen_port)

        s_ip4.set_property(NM.SETTING_IP_CONFIG_ROUTE_TABLE, fwmark)
        s_ip6.set_property(NM.SETTING_IP_CONFIG_ROUTE_TABLE, fwmark)
        w_con.set_property(NM.DEVICE_WIREGUARD_FWMARK, fwmark)
        w_con.set_property(NM.DEVICE_WIREGUARD_LISTEN_PORT, wg_listen_port)

        # The routing that is done by NM by default doesn't cut it
        # It automatically adds a suppress prefixlength rule such that LAN traffic is allowed
        # We want to make this configurable
        # Additionally, the overlap case with split tunnel doesn't work: https://codeberg.org/eduvpn/linux-app/issues/551

        prios = self.get_priorities(proxy is not None, allow_wg_lan)
        rules = [(4, AF_INET, s_ip4, 32, lan_4), (6, AF_INET6, s_ip6, 128, lan_6)]
        for ipver, family, setting, subnet, lans in rules:
            rule = NM.IPRoutingRule.new(family)
            rule.set_priority(prios[0])
            rule.set_invert(True)
            # fwmask 0xffffffff is the default
            rule.set_fwmark(fwmark, 0xFFFFFFFF)
            rule.set_table(fwmark)
            setting.add_routing_rule(rule)

            if proxy:
                dport_proxy = proxy.peer_port
                for proxy_peer_ip in proxy.peer_ips:
                    address = ip_address(proxy_peer_ip)
                    if address.version != ipver:
                        continue
                    proxy_rule = NM.IPRoutingRule.new(family)
                    proxy_rule.set_priority(prios[1])
                    sport = int(proxy.source_port)
                    proxy_rule.set_source_port(sport, sport)
                    proxy_rule.set_to(proxy_peer_ip, subnet)
                    proxy_rule.set_destination_port(dport_proxy, dport_proxy)
                    proxy_rule.set_ipproto(IPPROTO_TCP)
                    setting.add_routing_rule(proxy_rule)

            # when LAN should be allowed, we have to add a higher priority suppress prefixlength rule
            if allow_wg_lan:
                # make sure that traffic to the VPN ip range doesn't match the LAN rules
                exclude_ips = needs_exclude_v6
                if ipver == 4:
                    exclude_ips = needs_exclude_v4
                for ipaddr, subnet in exclude_ips:
                    exclude_rule = NM.IPRoutingRule.new(family)
                    exclude_rule.set_priority(prios[3])
                    exclude_rule.set_to(ipaddr, subnet)
                    exclude_rule.set_table(fwmark)
                    exclude_rule.set_suppress_prefixlength(0)
                    setting.add_routing_rule(exclude_rule)
                for lan in lans:
                    lan_rule = NM.IPRoutingRule.new(family)
                    lan_rule.set_to(str(lan.network_address), lan.prefixlen)
                    lan_rule.set_priority(prios[2])
                    lan_rule.set_suppress_prefixlength(0)
                    setting.add_routing_rule(lan_rule)

        w_con.append_peer(peer)
        private_key = config["Interface"]["PrivateKey"]
        w_con.set_property(NM.SETTING_WIREGUARD_PRIVATE_KEY, private_key)

        # set MTU if available in the config
        mtu = config["Interface"].get("MTU")
        if mtu:
            try:
                w_con.set_property(NM.SETTING_WIREGUARD_MTU, int(mtu))
            except ValueError:
                _logger.warning(f"got invalid WireGuard MTU value: {mtu}")

        profile.add_setting(s_ip4)
        profile.add_setting(s_ip6)
        profile.add_setting(s_con)
        profile.add_setting(w_con)

        self.proxy = proxy

        def set_callback(success: bool):
            if success and proxy is not None:
                self.proxy_tunnel(wg_listen_port)
            if callback is not None:
                callback(success)

        self.set_connection(profile, set_callback)  # type: ignore

    def new_cancellable(self):
        c = Cancellable.new()
        self.cancel_jar.add(c)
        c.connect(lambda: self.delete_cancellable(c))
        return c

    def new_connect_cancellable(self):
        c = Cancellable.new()
        c.connect(lambda: self.delete_connection_cancellable(c))
        self.cancel_jar.add(c)
        return c

    def delete_connection_cancellable(self, c):
        self.delete_cancellable(c)
        self.deactivate_connection()

    def cancel(self) -> bool:
        needed_cancel = len(self.cancel_jar.cookies) > 0
        self.cancel_jar.cancel()
        return needed_cancel

    def delete_cancellable(self, c):
        try:
            self.cancel_jar.delete(c)
        except ValueError:
            pass

    @run_in_glib_thread
    def activate_connection(self, callback: Optional[Callable] = None) -> None:
        con = self.client.get_connection_by_uuid(self.uuid)
        _logger.debug(f"activate_connection: {con}")
        if con is None:
            # Temporary workaround, connection is sometimes created too
            # late while according to the logging the connection is already
            # created. Need to find the correct event to sync on.
            time.sleep(0.1)
            GLib.idle_add(lambda: self.activate_connection(callback))
            return

        def activate_connection_callback(a_client, res, user_data=None):
            callback = None
            c = None
            # Add the ability to cancel the connecting...
            if user_data is not None:
                c, callback = user_data
            try:
                result = a_client.activate_connection_finish(res)
            except Exception as e:
                _logger.error(e)
                if c:
                    self.delete_cancellable(c)
                if callback:
                    callback(False)
            else:
                _logger.debug(f"activate_connection_async result: {result}")
                signal = None

                def changed_state(active: "NM.ActiveConnection", state_code: int, _reason_code: int):
                    state = NM.ActiveConnectionState(state_code)
                    if ConnectionState.from_active_state(state) == ConnectionState.CONNECTED:
                        if c:
                            self.delete_cancellable(c)
                        if signal:
                            active.disconnect(signal)
                        if callback:
                            callback(result is not None)

                    # This happens if the connection cancellable is called
                    if ConnectionState.from_active_state(state) == ConnectionState.DISCONNECTED:
                        if c:
                            self.delete_cancellable(c)
                        if signal:
                            active.disconnect(signal)
                        if callback:
                            callback(False)

                signal = self.active_connection.connect("state-changed", changed_state)

        c = self.new_connect_cancellable()
        self.client.activate_connection_async(
            connection=con,
            callback=activate_connection_callback,
            cancellable=c,
            user_data=(c, callback),
        )

    def deactivate_connection(self, callback: Optional[Callable] = None) -> None:
        connection = self.active_connection
        if connection is None:
            _logger.warning(f"no connection to deactivate of uuid {uuid}")
            if callback:
                callback(False)
            return
        type = connection.get_connection_type()
        if type == "vpn":
            self.deactivate_connection_vpn(callback)
        elif type == "wireguard":
            self.deactivate_connection_wg(callback)
        else:
            _logger.warning(f"unexpected connection type {type}")
            if callback:
                callback(False)

    @run_in_glib_thread
    def deactivate_connection_vpn(self, callback: Optional[Callable] = None) -> None:
        con = self.active_connection
        _logger.debug(f"deactivate_connection uuid: {uuid} connection: {con}")
        if con:

            def on_deactivate_connection(a_client: "NM.Client", res, user_data=None):
                callback = None
                c = None
                if user_data is not None:
                    c, callback = user_data
                try:
                    result = a_client.deactivate_connection_finish(res)
                except Exception as e:
                    _logger.error(f"deactive_connection_async exception: {e}")
                    if callback:
                        callback(False)
                else:
                    _logger.debug(f"deactivate_connection_async result: {result}")
                finally:

                    def on_deleted(success: bool):
                        if c:
                            self.delete_cancellable(c)
                        if callback:
                            # Whether or not deletion was a success, we return true
                            callback(success)

                    self.delete_connection(on_deleted)

            c = self.new_cancellable()
            self.client.deactivate_connection_async(
                active=con,
                callback=on_deactivate_connection,
                cancellable=c,
                user_data=(c, callback),
            )
        else:
            _logger.debug("No active connection to deactivate")

    @run_in_glib_thread
    def delete_connection(self, callback: Callable) -> None:
        # We run the disconnected callback early if a delete fail happens
        if self.uuid is None:
            _logger.debug("No uuid found for deleting the connection")
            callback(False)
            return

        con = self.client.get_connection_by_uuid(self.uuid)
        if con is None:
            _logger.debug(f"No uuid connection found to delete with uuid {uuid}")
            callback(False)
            return

        # Delete the connection and after that do the callback
        def on_deleted(a_con: "NM.RemoteConnection", res, user_data=None):
            callback = None
            c = None
            if user_data is not None:
                c, callback = user_data
            try:
                result = a_con.delete_finish(res)
            except Exception as e:
                _logger.error(f"delete_connection_async exception: {e}")
                if callback:
                    callback(False)
            else:
                _logger.debug(f"delete_async result: {result}")
            finally:
                if c:
                    self.delete_cancellable(c)
                if callback:
                    callback(True)

        c = self.new_cancellable()
        con.delete_async(callback=on_deleted, cancellable=c, user_data=(c, callback))

    @property
    def wireguard_device(self) -> Optional["NM.DeviceWireGuard"]:
        devices = [
            device
            for device in self.client.get_all_devices()
            if device.get_type_description() == "wireguard"
            and self.uuid in {conn.get_uuid() for conn in device.get_available_connections()}
        ]
        if not devices:
            return None

        return devices[0]

    @run_in_glib_thread
    def deactivate_connection_wg(self, callback: Optional[Callable] = None) -> None:
        def on_disconnect(a_device: "NM.DeviceWireGuard", res, user_data=None):
            callback = None
            c = None
            if user_data is not None:
                c, callback = user_data
            try:
                result = a_device.disconnect_finish(res)
            except Exception as e:
                _logger.error(e)
                if callback:
                    callback(False)
            else:
                _logger.debug(f"disconnect_async result: {result}")
            finally:

                def on_deleted(success: bool):
                    if c:
                        self.delete_cancellable(c)
                    if callback:
                        callback(success)

                self.delete_connection(on_deleted)

        _logger.debug(f"disconnect uuid: {uuid}")
        device = self.wireguard_device
        if device is None:
            _logger.warning("Cannot disconnect, no WireGuard device")
            return
        c = self.new_cancellable()
        device.disconnect_async(callback=on_disconnect, cancellable=c, user_data=(c, callback))

    @run_in_background_thread("restart-proxy")
    def restart_proxy(self, conn):
        if self.proxy is None:
            return
        signal = None

        def changed_state(active: "NM.ActiveConnection", state_code: int, _reason_code: int):
            state = NM.ActiveConnectionState(state_code)
            if ConnectionState.from_active_state(state) != ConnectionState.CONNECTED:
                return
            time.sleep(1)
            if self.proxy is not None:
                self.proxy.restart()
            if signal is not None:
                active.disconnect(signal)

        signal = conn.connect("state-changed", changed_state)

    def subscribe_to_status_changes(
        self,
        callback: Callable[[ConnectionState], Any],
    ) -> bool:
        """
        Subscribe to network status changes via the NM client.

        The callback argument is called with the connection state and reason
        whenever they change.
        """

        # The callback to monitor state changes
        # Let the state machine know for state updates
        def wrapped_callback(active: "NM.ActiveConnection", state_code: int, _reason_code: int):
            if active.get_uuid() != self.uuid:
                return

            state = NM.ActiveConnectionState(state_code)

            if ConnectionState.from_active_state(state) == ConnectionState.DISCONNECTED:
                self.proxy = None

            callback(ConnectionState.from_active_state(state))

        # Connect the state changed signal for an active connection
        def connect(con: "NM.ActiveConnection"):
            con.connect("state-changed", wrapped_callback)

        # The callback when a connection gets added
        # Connect the signals
        def wrapped_connection_added(client: "NM.Client", active_con: "NM.ActiveConnection"):
            if active_con.get_uuid() != self.uuid:
                self.restart_proxy(active_con)
                return
            connect(active_con)

        # If a connection was found already then...
        active_con = self.active_connection

        if active_con:
            connect(active_con)

        # Connect the active connection added signal
        self.client.connect("active-connection-added", wrapped_connection_added)
        return True


def action_with_mainloop(action: Callable):
    _logger.debug("calling action with CLI mainloop")
    main_loop = GLib.MainLoop()

    def quit_loop(*args, **kwargs):
        _logger.debug("Quiting main loop, thanks!")
        main_loop.quit()

    # Schedule the action
    GLib.idle_add(lambda: action(callback=quit_loop))

    # Run the main loop
    main_loop.run()


def add_connection_callback(client: "NM.Client", result: Task, user_data) -> None:
    object, c, callback = user_data
    try:
        new_con = client.add_connection_finish(result)
    except Exception as e:
        object.delete_cancellable(c)
        _logger.error(f"add connection error: {e}")
        if callback is not None:
            callback(False)
    else:
        object.delete_cancellable(c)
        object.uuid = new_con.get_uuid()
        _logger.debug(f"Connection added for uuid: {object.uuid}")
        if callback is not None:
            callback(new_con is not None)



import json
import logging
import os
from abc import ABC, abstractmethod

import gi

secureKeyring = True
try:
    gi.require_version("Secret", "1")
    from gi.repository import Secret  # type: ignore
except (ValueError, ImportError):
    secureKeyring = False


logger = logging.getLogger(__name__)


class TokenKeyring(ABC):
    def __init__(self, variant):
        self.variant = variant

    @property
    def available(self) -> bool:
        return True

    @property
    def secure(self) -> bool:
        return False

    @abstractmethod
    def clear(self, attributes) -> bool:
        pass

    @abstractmethod
    def save(self, label, attributes, secret):
        pass

    @abstractmethod
    def load(self, attributes):
        pass


class DBusKeyring(TokenKeyring):
    """A keyring using libsecret with DBus"""

    def __init__(self, variant):
        super().__init__(variant)
        # None is the default collection
        self.collection = None

    # This keyring is secure
    def secure(self):
        return True

    @property
    def available(self):
        # If import was not successful, this is definitely not available
        if not secureKeyring:
            logger.warning("keyring not available due to libsecret import not available")
            return False

        # Libs available, do a test run
        try:
            attributes = {"test": "test"}
            secret = "eduVPN test"
            self.save("eduVPN testing run", attributes, secret)
            assert self.load(attributes) == secret
            self.clear(attributes)
        except (gi.repository.GLib.Error, AssertionError) as e:
            logger.warning(f"error when checking if keyring is available: {e}")
            return False
        return True

    def create_schema(self, attributes):
        return Secret.Schema.new(
            self.variant.name,
            Secret.SchemaFlags.NONE,
            {k: Secret.SchemaAttributeType.STRING for k in attributes},
        )

    def clear(self, attributes) -> bool:
        schema = self.create_schema(attributes)
        return Secret.password_clear_sync(schema, attributes, None)

    def save(self, label, attributes, secret):
        # Prefix the label with the client name
        label = f"{self.variant.name} - {label}"
        schema = self.create_schema(attributes)
        return Secret.password_store_sync(schema, attributes, self.collection, label, str(secret), None)

    def load(self, attributes):
        """Load a password in the secret service, return None when found nothing"""
        schema = self.create_schema(attributes)
        return Secret.password_lookup_sync(schema, attributes, None)


JSON_VERSION = "v1"


class InsecureFileKeyring(TokenKeyring):
    def __init__(self, variant):
        super().__init__(variant)

    @property
    def filename(self):
        return self.variant.config_prefix / "keys"

    def unique_key(self, attributes):
        return ",".join(list(attributes.values()))

    def load_previous(self):
        with open(self.filename, "r") as f:
            try:
                c = json.load(f)
            except Exception as e:
                logger.debug(f"failed to load JSON: {str(e)}")
                c = {}
            return c.get(JSON_VERSION, {})

    def write(self, vals):
        towrite = {JSON_VERSION: vals}
        with open(self.filename, "w+") as f:
            json.dump(towrite, f)

    def clear(self, attributes) -> bool:
        # Get previous entries
        new = {}
        if os.path.exists(self.filename):
            new = self.load_previous()
        key = self.unique_key(attributes)
        new.pop(key, None)
        self.write(new)
        return True

    def save(self, label, attributes, secret):
        new = {}

        # Get previous entries
        if os.path.exists(self.filename):
            new = self.load_previous()

        # add/overwrite new entry
        key = self.unique_key(attributes)
        new[key] = secret

        # Write new values
        self.write(new)

    def load(self, attributes):
        if not os.path.exists(self.filename):
            return None
        previous = self.load_previous()
        key = self.unique_key(attributes)
        return previous.get(key, None)



import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TextIO
from unittest import TestCase
from unittest.mock import PropertyMock, patch

from eduvpn.nm import NMManager
from eduvpn.ui import stats
from eduvpn.variants import EDUVPN

MOCK_IFACE = "mock"


def write_temp_stats_file(filepath: Path, total_bytes: int):
    f = open(filepath, "w")
    f.write(str(total_bytes))
    f.close()


def try_open(path: Path):
    if not path.is_file():
        return None
    return open(path, "r")


@patch("eduvpn.nm.NMManager.iface", new_callable=PropertyMock, return_value=MOCK_IFACE)
class TestStats(TestCase):
    def test_stat_bytes(self, _):
        nm_manager = NMManager(EDUVPN, None)
        with TemporaryDirectory() as tempdir:
            # Create test data in the wanted files
            # Use the tempdir so it is cleaned up later
            values = [0, 37, 6166746255814, -43]
            # For the expected values we want the human readable string
            # The last two are special
            #   - 0 B because the value has decreased
            #   - default text because the file does not exist
            expected_values = ["0 B", "37 B", "5.61 TB", "0 B"]

            # Create the statistics path
            stat_path = Path(tempdir) / MOCK_IFACE / "statistics"
            os.makedirs(stat_path)

            # Create the download and upload files
            download_file = stat_path / "rx_bytes"
            write_temp_stats_file(download_file, "0")
            download_filehandler = try_open(download_file)

            upload_file = stat_path / "tx_bytes"
            write_temp_stats_file(upload_file, "0")
            upload_filehandler = try_open(upload_file)

            def check_expected(_property: str, _file: TextIO):
                # Create the class instance
                class_ = stats.NetworkStats(nm_manager)

                # Loop over the files,
                # patch it and check if the expected value holds
                class_ = stats.NetworkStats(nm_manager)
                for i, expected in enumerate(expected_values):
                    write_temp_stats_file(Path(_file.name), values[i])
                    self.assertEqual(getattr(class_, _property), expected)
                _file.close()

            # Mock the files and check the expected values
            with patch(
                "eduvpn.ui.stats.NetworkStats.upload_file",
                new_callable=PropertyMock,
                return_value=upload_filehandler,
            ):
                check_expected("upload", upload_filehandler)
            with patch(
                "eduvpn.ui.stats.NetworkStats.download_file",
                new_callable=PropertyMock,
                return_value=download_filehandler,
            ):
                check_expected("download", download_filehandler)



import json
import logging
import os
import signal
import time
import webbrowser
from typing import Any, Callable, Iterator, Optional, TextIO

from eduvpn_common.main import EduVPN, ServerType, WrappedError
from eduvpn_common.state import State, StateType
from eduvpn_common.types import ProxySetup, ReadRxBytes  # type: ignore[attr-defined]

from eduvpn import nm
from eduvpn.config import Configuration
from eduvpn.connection import (
    Config,
    Connection,
    Protocol,
    parse_config,
    parse_expiry,
    parse_tokens,
)
from eduvpn.keyring import DBusKeyring, InsecureFileKeyring, TokenKeyring
from eduvpn.proxy import Proxy
from eduvpn.server import ServerDatabase, parse_profiles, parse_required_transition
from eduvpn.utils import (
    handle_exception,
    model_transition,
    run_in_background_thread,
    run_in_glib_thread,
    set_failovered,
    set_online_detecting,
    set_server_list_refresh,
)
from eduvpn.variants import ApplicationVariant

logger = logging.getLogger(__name__)


class ApplicationModelTransitions:
    def __init__(self, common: EduVPN, variant: ApplicationVariant) -> None:
        self.common = common
        self.common.register_class_callbacks(self)
        self.server_db = ServerDatabase(common, variant.use_predefined_servers)

    @model_transition(State.MAIN, StateType.ENTER)
    def get_previous_servers(self, old_state: State, data):
        logger.debug(f"Transition: MAIN, old state: {old_state}")
        return self.server_db.configured

    @model_transition(State.DEREGISTERED, StateType.ENTER)
    def deregistered(self, old_state: State, data: str):
        logger.debug(f"Transition: DEREGISTERED, old state: {old_state}")
        return data

    @model_transition(State.ADDING_SERVER, StateType.ENTER)
    def adding_server(self, old_state: State, data: str):
        logger.debug(f"Transition: ADDING_SERVER, old state: {old_state}")
        return data

    @model_transition(State.GETTING_CONFIG, StateType.ENTER)
    def getting_config(self, old_state: State, data: str):
        logger.debug(f"Transition: GETTING_CONFIG, old state: {old_state}")
        return data

    @model_transition(State.DISCONNECTED, StateType.ENTER)
    def disconnected_server(self, old_state: State, data: str):
        logger.debug(f"Transition: DISCONNECTED, old state: {old_state}")
        return self.server_db.current

    @model_transition(State.DISCONNECTING, StateType.ENTER)
    def disconnecting(self, old_state: State, data):
        logger.debug(f"Transition: DISCONNECTING, old state: {old_state}")
        return self.server_db.current

    @model_transition(State.ASK_PROFILE, StateType.ENTER)
    def ask_profile(self, old_state: State, data: str):
        logger.debug(f"Transition: ASK_PROFILE, old state: {old_state}")
        cookie, profiles = parse_required_transition(data, get=parse_profiles)

        def set_profile(prof):
            self.common.cookie_reply(cookie, prof)

        return (set_profile, profiles)

    @model_transition(State.ASK_LOCATION, StateType.ENTER)
    def ask_location(self, old_state: State, data):
        cookie, locations = parse_required_transition(data)

        def set_location(loc):
            self.common.cookie_reply(cookie, loc)

        return (set_location, locations)

    @model_transition(State.OAUTH_STARTED, StateType.ENTER)
    def start_oauth(self, old_state: State, url: str):
        logger.debug(f"Transition: OAUTH_STARTED, old state: {old_state}")
        self.open_browser(url)
        return url

    @model_transition(State.GOT_CONFIG, StateType.ENTER)
    def parse_config(self, old_state: State, data):
        logger.debug(f"Transition: GOT_CONFIG, old state: {old_state}")
        return self.server_db.current

    @run_in_background_thread("open-browser")
    def open_browser(self, url):
        logger.debug(f"Opening web browser with url: {url}")
        webbrowser.open(url)
        # Explicitly wait to not have zombie processes
        # See https://bugs.python.org/issue5993
        logger.debug("Running os.wait for browser")
        try:
            os.wait()
        except ChildProcessError:
            pass
        logger.debug("Done waiting for browser")

    @model_transition(State.CONNECTED, StateType.ENTER)
    def parse_connected(self, old_state: State, data):
        logger.debug(f"Transition: CONNECTED, old state: {old_state}")
        server = self.server_db.current
        expire_times = parse_expiry(self.common.get_expiry_times())
        return (server, expire_times)

    @model_transition(State.CONNECTING, StateType.ENTER)
    def parse_connecting(self, old_state: State, data):
        logger.debug(f"Transition: CONNECTING, old state: {old_state}")
        return self.server_db.current


class ApplicationModel:
    def __init__(
        self,
        common: EduVPN,
        config,
        variant: ApplicationVariant,
        nm_manager,
    ) -> None:
        self.common = common
        self.config = config
        self._keyring: TokenKeyring | None = None
        self.transitions = ApplicationModelTransitions(common, variant)
        self.variant = variant
        self.nm_manager = nm_manager
        self._was_tcp = False
        self._should_failover = False
        self._proxy_setup_handler = ProxySetup(self.on_proxy_setup)

    @property
    def keyring(self):
        if self._keyring is not None:
            return self._keyring
        keyring = DBusKeyring(self.variant)
        if not keyring.available:
            logger.warning("Secure keyring not available, reverting to insecure file keyring!")
            keyring = InsecureFileKeyring(self.variant)
        self._keyring = keyring
        return self._keyring

    def refresh_list(self):
        set_server_list_refresh(self.common, self.server_db.configured)

    def register(self):
        self.common.register()
        self.common.set_token_handler(self.load_tokens, self.save_tokens)

    def cancel(self):
        # Cancel any eduvpn-common operation
        self.common.cancel()

        # Cancel any NetworkManager operation
        self.nm_manager.cancel()

    @property
    def server_db(self):
        return self.transitions.server_db

    @property
    def current_server(self):
        return self.server_db.current

    def get_failover_rx(self, filehandler: Optional[TextIO]) -> int:
        rx_bytes = self.nm_manager.get_stats_bytes(filehandler)
        if rx_bytes is None:
            return -1
        return rx_bytes

    def should_failover(self):
        if self._should_failover:
            logger.debug("eduvpn-common reports we should failover")

            if self._was_tcp:
                logger.debug("TCP was not previously triggered, failover should not continue")
                return False
            return True

        logger.debug("Failover should not continue")
        return False

    def reconnect_tcp(self, callback: Callable):
        def on_reconnected(success: bool):
            callback(success)

        self.reconnect(on_reconnected, prefer_tcp=True)

    @run_in_background_thread("start-failover")
    def start_failover(self, callback: Callable):
        try:
            rx_bytes_file = self.nm_manager.open_stats_file("rx_bytes")
            if rx_bytes_file is None:
                logger.error("Failed to initialize failover, failed to open rx bytes file")
                callback(False)
                return
            endpoint = self.nm_manager.failover_endpoint_ip
            if endpoint is None:
                logger.error("Failed to initialize failover, failed to get endpoint")
                callback(False)
                return
            mtu = self.nm_manager.mtu
            if mtu is None:
                logger.debug("failed to get MTU for failover, setting MTU to 1000")
                mtu = 1000
            logger.debug(
                f"starting failover with gateway {endpoint} and MTU {mtu} for protocol {self.nm_manager.protocol}"
            )
            failover_delay = float(os.getenv("EDUVPN_FAILOVER_DELAY", 1))
            logger.debug(f"Sleeping for {failover_delay}s to begin failover")
            time.sleep(failover_delay)
            dropped = self.common.start_failover(
                endpoint,
                mtu,
                ReadRxBytes(lambda: self.get_failover_rx(rx_bytes_file)),
            )

            if dropped:
                logger.debug("Failover exited, connection is dropped")
                callback(True)
            else:
                logger.debug("Failover exited, connection is NOT dropped")
                callback(False)
        except WrappedError as e:
            logger.debug(f"Failed failover, error: {e}")
            callback(False)

    def change_secure_location(self, country_code: str):
        server = self.server_db.secure_internet
        if server.country_code == country_code:
            return
        self.common.set_secure_location(server.org_id, country_code)
        self.common.set_state(State.MAIN)

    def go_back(self):
        self.cancel()
        self.common.set_state(State.MAIN)

    def delisted_disco(self):
        self.server_db.disco_update(search="", cache=False)

        # refresh the server list if we're still in the main state
        if self.common.in_state(State.MAIN):
            self.refresh_list()

    def add(self, server, callback=None):
        self.common.add_server(server.category_id, server.identifier)
        if callback:
            callback(server)

    def remove(self, server):
        self.common.remove_server(server.category_id, server.identifier)
        # Delete tokens from the keyring
        self.clear_tokens(server.category_id, server.identifier)
        self.common.set_state(State.MAIN)

    def connect_get_config(self, server, prefer_tcp: bool = False) -> Config:
        # We prefer TCP if the user has set it or UDP is determined to be blocked
        config = self.common.get_config(server.category_id, server.identifier, prefer_tcp)
        return parse_config(config)

    def clear_tokens(self, server_type: int, server_id: str):
        attributes = {
            "server": server_id,
            "category": str(ServerType(server_type)),
        }
        try:
            cleared = self.keyring.clear(attributes)
            if not cleared:
                logger.debug("Tokens were not cleared")
        except Exception as e:
            logger.debug("Failed clearing tokens with exception")
            logger.debug(e, exc_info=True)

    def load_tokens(self, server_id: str, server_type: int) -> Optional[str]:
        attributes = {"server": server_id, "category": str(ServerType(server_type))}
        try:
            tokens_json = self.keyring.load(attributes)
            if tokens_json is None:
                logger.debug("No tokens available")
                return None
            tokens = json.loads(tokens_json)
            expires = tokens.get("expires_at", None)
            if expires is None:
                expires = tokens.get("expires", None)
            if expires is None:
                logger.warning("failed to parse expires")
                return None
            d = {
                "access_token": tokens["access"],
                "refresh_token": tokens["refresh"],
                "expires_at": int(expires),
            }
            return json.dumps(d)
        except Exception as e:
            logger.debug("Failed loading tokens with exception:")
            logger.debug(e, exc_info=True)
            return None

    def save_tokens(self, server_id: str, server_type: int, tokens: str):
        tokens_parsed = parse_tokens(tokens)
        if tokens is None or (tokens_parsed.access == "" and tokens_parsed.refresh == ""):
            logger.warning("Got empty tokens, not saving them to the keyring")
            return
        tokens_dict = {}
        tokens_dict["access"] = tokens_parsed.access
        tokens_dict["refresh"] = tokens_parsed.refresh
        tokens_dict["expires_at"] = str(tokens_parsed.expires)
        attributes = {"server": server_id, "category": str(ServerType(server_type))}
        label = f"{server_id} - OAuth Tokens"
        try:
            self.keyring.save(label, attributes, json.dumps(tokens_dict))
        except Exception as e:
            logger.error("Failed saving tokens with exception:")
            logger.error(e, exc_info=True)

    def on_proxy_setup(self, fd):
        logger.debug(f"got proxy fd: {fd}")

    def connect(
        self,
        server,
        callback: Optional[Callable] = None,
        prefer_tcp: bool = False,
    ) -> None:
        # Variable to be used as a last resort or for debugging
        # to override the prefer TCP setting
        if os.environ.get("EDUVPN_PREFER_TCP", "0") == "1":
            prefer_tcp = True
        config = self.connect_get_config(server, prefer_tcp=prefer_tcp)
        if not config:
            logger.warning("no configuration available")
            if callback:
                callback(False)
            return

        self._was_tcp = prefer_tcp or config.protocol == Protocol.WIREGUARDTCP
        self._should_failover = config.should_failover

        def on_fail():
            if self.common.in_state(State.CONNECTED):
                self.common.set_state(State.DISCONNECTING)
                self.common.set_state(State.DISCONNECTED)
            if callback:
                callback(False)

        def on_success():
            # failed to disconnected
            if self.common.in_state(State.CONNECTING):
                self.common.set_state(State.CONNECTED)
            if callback:
                callback(True)

        def final_connected(dropped: bool, error: str = ""):
            # failover reports not dropped, return to connected if we are still in connecting
            if not dropped:
                if self.common.in_state(State.CONNECTING):
                    on_success()
                else:
                    on_fail()
                return

            # Connection is dropped!
            def on_reconnected(success: bool):
                if not success:
                    handle_exception(self.common, Exception("failed to reconnect with TCP"))
                if success:
                    set_failovered(self.common)
                # When an error happens we always set success now as the previous
                # protocol might be connected to
                # TODO: differentiate between disconnect and new connect errors
                on_success()

            # reconnect with TCP
            self.reconnect_tcp(on_reconnected)

        def on_connected(success: bool):
            if success:
                # failover should not continue
                if not self.should_failover():
                    on_success()
                else:
                    set_online_detecting(self.common)
                    # failover should start
                    self.start_failover(final_connected)
            else:
                on_fail()

        def on_connect(success: bool):
            if success:
                self.nm_manager.activate_connection(on_connected)
            else:
                on_fail()

        @run_in_glib_thread
        def connect(config):
            if not self.common.in_state(State.CONNECTING):
                self.common.set_state(State.CONNECTING)
            connection = Connection.parse(config)
            proxy = None
            if config.protocol is Protocol.WIREGUARDTCP:
                proxy = Proxy(self.common, connection.proxy_endpoint)
                wrapper_proxy = self.common.new_proxyguard(
                    proxy.listen_port, proxy.source_port, proxy.endpoint, self._proxy_setup_handler
                )
                proxy.wrapper = wrapper_proxy
            connection.connect(
                self.nm_manager,
                config.default_gateway,
                self.config.allow_wg_lan,
                config.dns_search_domains,
                proxy,
                on_connect,
            )

        connect(config)

    def reconnect(self, callback: Optional[Callable] = None, prefer_tcp: bool = False):
        def on_disconnected(success: bool):
            if success:
                self.activate_connection(callback, prefer_tcp=prefer_tcp)

        # Reconnect
        self.deactivate_connection(on_disconnected)

    # https://github.com/eduvpn/documentation/blob/v3/API.md#session-expiry
    def renew_session(self, callback: Optional[Callable] = None):
        was_connected = self.common.in_state(State.CONNECTED)

        @run_in_background_thread("reconnect")
        def reconnect(success: bool = True):
            if not success:
                if callback:
                    callback(False)
                return
            # Delete the OAuth access and refresh token
            # Start the OAuth authorization flow
            self.common.renew_session()
            # Automatically reconnect to the server
            self.activate_connection(callback)

        if was_connected:
            # Call /disconnect and reconnect with callback
            self.deactivate_connection(reconnect)
        else:
            reconnect()

    def disconnect(self, callback: Optional[Callable] = None) -> None:
        self.nm_manager.deactivate_connection(callback)

    def set_profile(self, profile: str, connect=False):
        was_connected = self.common.in_state(State.CONNECTED)

        def do_profile(success: bool = True):
            if not success:
                return
            # Set the profile ID
            self.common.set_profile(profile)

            # Connect if we should and if we were previously connected
            if connect and was_connected:
                self.activate_connection()

        # Deactivate connection if we are connected
        # and the connection should be modified
        # the do_profile will be called in the callback
        if was_connected and connect:
            self.deactivate_connection(do_profile)
        else:
            do_profile()

    def activate_connection(self, callback: Optional[Callable] = None, prefer_tcp: bool = False):
        if (
            not self.common.in_state(State.GOT_CONFIG)
            and not self.common.in_state(State.DISCONNECTED)
            and not self.common.in_state(State.MAIN)
        ):
            if callback:
                logger.error("invalid state to activate connection")
                callback(False)
            return
        if not self.current_server:
            if callback:
                logger.error("failed to get current server")
                callback(False)
            return

        def on_connected(success: bool):
            if callback:
                callback(success)

        self.connect(self.current_server, on_connected, prefer_tcp=prefer_tcp)

    @run_in_background_thread("cleanup")
    def cleanup(self, callback: Optional[Callable] = None):
        # We retry this cleanup 2 times
        retries = 2

        # Try to cleanup with a number of retries
        for i in range(retries):
            logger.debug("Cleaning up tokens...")
            try:
                self.common.cleanup()
            except Exception as e:
                # We can try again
                if i < retries - 1:
                    logger.debug(
                        f"Got an error: {str(e)} while cleaning up, try number: {i + 1}. This could mean the connection was not fully disconnected yet. Trying again..."
                    )
                else:
                    # All retries are done
                    logger.debug(f"Got an error: {str(e)} while cleaning up, after full retries: {i + 1}.")
            else:
                break
        if self.common.in_state(State.DISCONNECTING):
            self.common.set_state(State.DISCONNECTED)
        if callback:
            callback()

    def deactivate_connection(self, callback: Optional[Callable] = None, cleanup=True) -> None:
        curr = None
        if self.common.in_state(State.CONNECTED):
            curr = State.CONNECTED

        if self.common.in_state(State.CONNECTING):
            curr = State.CONNECTING

        if not curr:
            return
        self.common.set_state(State.DISCONNECTING)

        def on_disconnected(success: bool):
            if success:

                def on_cleaned():
                    self.common.cancel()
                    if callback:
                        callback(True)

                if cleanup:
                    self.cleanup(on_cleaned)
                else:
                    if self.common.in_state(State.DISCONNECTING):
                        self.common.set_state(State.DISCONNECTED)
                    on_cleaned()
            else:
                try:
                    self.common.set_state(curr)
                except WrappedError as e:
                    logger.debug(f"set_state error in deactivate connection: {str(e)}")
                if callback:
                    callback(False)

        self.disconnect(on_disconnected)

    def search_predefined(self, query: str) -> Iterator[Any]:
        return self.server_db.search_predefined(query)

    def search_custom(self, query: str) -> Iterator[Any]:
        return self.server_db.search_custom(query)


class Application:
    def __init__(self, variant: ApplicationVariant, common: EduVPN) -> None:
        self.variant = variant
        self.nm_manager = nm.NMManager(variant, common)
        self.common = common
        directory = variant.config_prefix
        self.config = Configuration.load(directory)
        self.model = ApplicationModel(common, self.config, variant, self.nm_manager)

        signal.signal(signal.SIGINT, lambda s, f: self.cleanup_sigint(s, f))

    def cleanup_sigint(self, _signal, _frame):
        self.model.cancel()
        self.common.deregister()

    def on_network_update_callback(self, state, initial=False):
        try:
            if state == nm.ConnectionState.CONNECTED:
                if (
                    not self.common.in_state(State.DISCONNECTED)
                    and not self.common.in_state(State.GOT_CONFIG)
                    and not initial
                ):
                    return
                if not initial:
                    self.common.set_state(State.CONNECTING)
                # Already connected
                self.common.set_state(State.CONNECTED)
            elif state == nm.ConnectionState.CONNECTING:
                self.common.set_state(State.CONNECTING)
            elif state == nm.ConnectionState.DISCONNECTED:
                if not self.common.in_state(State.CONNECTED) and not self.common.in_state(State.CONNECTING):
                    return
                self.common.set_state(State.DISCONNECTING)
                self.model.cancel()
                self.model.cleanup()
                self.model.disconnect()
        except Exception as e:
            logger.debug(f"error occurred: {str(e)}")
            return

    def initialize_network(self, needs_update=True) -> None:
        """
        Determine the current network state.
        """
        # Check if a previous network configuration exists.
        uuid = self.nm_manager.existing_connection
        if uuid:
            self.on_network_update_callback(self.nm_manager.connection_state, needs_update)

        @run_in_background_thread("on-network-update")
        def update(state):
            self.on_network_update_callback(state, False)

        self.nm_manager.subscribe_to_status_changes(update)



import logging
import os
import sys
import threading
import traceback
from functools import lru_cache, partial, wraps
from gettext import gettext
from os import environ, path
from sys import prefix
from typing import Callable, Optional, Union

from eduvpn_common.event import class_state_transition
from eduvpn_common.main import WrappedError
from eduvpn_common.state import State, StateType

logger = logging.getLogger(__file__)


def get_logger(name_space: str) -> logging.Logger:
    return logging.getLogger(name_space)


def handle_exceptions(exc_type, exc_value, exc_traceback):
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return

    logger.error("Unhandled exception", exc_info=(exc_type, exc_value, exc_traceback))


def get_ui_state(state: State) -> int:
    # The UI state will have as identifier the last state id + offset of the state
    # So for example the UI DEREGISTERED state will come directly after the last normal state
    return len(State) + state


ERROR_STATE = 2 * len(State) + 1
ONLINEDETECT_STATE = ERROR_STATE + 1
FAILOVERED_STATE = ERROR_STATE + 2
SERVER_LIST_REFRESH_STATE = ERROR_STATE + 3


def handle_exception(common, exception):
    log_exception(exception)
    common.event_handler.run(get_ui_state(ERROR_STATE), get_ui_state(ERROR_STATE), exception)


def set_online_detecting(common):
    common.event_handler.run(get_ui_state(ONLINEDETECT_STATE), get_ui_state(ONLINEDETECT_STATE), "")


def set_failovered(common):
    common.event_handler.run(get_ui_state(FAILOVERED_STATE), get_ui_state(FAILOVERED_STATE), "")


def set_server_list_refresh(common, servers):
    common.event_handler.run(
        get_ui_state(SERVER_LIST_REFRESH_STATE),
        get_ui_state(SERVER_LIST_REFRESH_STATE),
        servers,
    )


def model_transition(state: State, state_type: StateType) -> Callable:
    def decorator(func):
        @run_in_background_thread(str(func))
        def inner(self, other_state, data):
            # The model converts the data
            try:
                model_converted = func(self, other_state, data)
            except Exception as e:
                handle_exception(self.common, e)
                return

            other_ui_state = get_ui_state(other_state)
            ui_state = get_ui_state(state)
            # We can then pass it to the UI
            if state_type == StateType.ENTER:
                self.common.event_handler.run(other_ui_state, ui_state, model_converted)
            else:
                self.common.event_handler.run(ui_state, other_ui_state, model_converted)

        # Add the inner function on the state transition
        class_state_transition(state, state_type)(inner)

        # Return the inner function to be called
        return inner

    return decorator


def ui_transition(state: State, state_type: StateType) -> Callable:
    def decorator(func):
        @run_in_glib_thread
        @class_state_transition(get_ui_state(state), state_type)
        def inner(self, other_state, data):
            func(self, other_state, data)

        return inner

    return decorator


def cmd_transition(state: State, state_type: StateType):
    def decorator(func):
        @class_state_transition(get_ui_state(state), state_type)
        def inner(self, other_state, data):
            func(self, other_state, data)

        return inner

    return decorator


def init_logger(debug: bool, logfile, mode):
    log_format = "%(asctime)s - %(threadName)s - %(levelname)s - %(name)s - %(filename)s:%(lineno)d - %(message)s"
    os.makedirs(
        os.path.dirname(logfile),
        mode=mode,
        exist_ok=True,
    )
    if debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO
    logging.basicConfig(
        level=log_level,
        format=log_format,
        handlers=[
            logging.FileHandler(logfile),
            logging.StreamHandler(),
        ],
    )
    # Log unhandled exceptions
    sys.excepthook = handle_exceptions


def log_exception(exception: Exception):
    if isinstance(exception, WrappedError) and exception.misc:
        logger.debug(f"eduvpn-common misc error returned: {str(exception)}")
    else:
        # Other exceptions are already logged by Go
        logger.error(f"exception occurred: {str(exception)}")
        traceback.print_exc()


@lru_cache(maxsize=1)
def get_prefix() -> str:
    """
    Returns the Python prefix where eduVPN is installed

    returns:
        path to Python installation prefix
    """
    target = "share/eduvpn/builder/mainwindow.ui"
    local = f"{path.dirname(path.abspath(path.abspath(__file__)))}/data"
    options = [local, path.expanduser("~/.local"), "/usr/local", prefix]
    for option in options:
        logger.debug(f"looking for '{target}' in '{option}'")
        if path.isfile(path.join(option, target)):
            return option
    raise Exception("Can't find eduVPN installation")


def get_config_dir() -> str:
    return environ.get("XDG_CONFIG_HOME", "~/.config")


def thread_helper(func: Callable, *, name: Optional[str] = None) -> threading.Thread:
    """
    Runs a function in a thread

    args:
        func (lambda): a function to run in the background
    """
    thread = threading.Thread(target=func, name=name)
    thread.daemon = True
    thread.start()
    return thread


def run_in_background_thread(name: Optional[str] = None) -> Callable:
    """
    Decorator for functions that must always run
    in a background thread.
    """

    def decorator(func):
        @wraps(func)
        def background_func(*args, **kwargs):
            thread_helper(partial(func, *args, **kwargs), name=name)

        return background_func

    return decorator


def run_in_glib_thread(func: Union[partial, Callable]) -> Callable:
    """
    Decorator for functions that must always run
    in the main GTK thread.
    """
    from gi.repository import GLib

    @wraps(func)
    def main_gtk_thread_func(*args, **kwargs):
        GLib.idle_add(partial(func, *args, **kwargs))

    return main_gtk_thread_func


def run_periodically(
    func: Callable[[], None],
    interval: float,
    name: Optional[str] = None,
) -> Callable[[], None]:
    """
    Run a funtion periodically in a background thread.

    The given function is called every `interval` seconds,
    until either it returns False
    or until the returned cancel callback is called.
    """
    if name is None:
        name = "run-periodically"
    event = threading.Event()

    @run_in_background_thread(name)
    def run_periodic_thread():
        if func() is False:
            return

        while 1:
            if event.wait(interval):
                return
            elif func() is False:
                return

    run_periodic_thread()
    return event.set


def get_human_readable_bytes(total_bytes: int) -> str:
    """
    Helper function to calculate the human readable bytes.
    E.g. B, kB, MB, GB, TB.
    """
    suffix = ""
    hr_bytes = float(total_bytes)
    for suffix in ["B", "kB", "MB", "GB", "TB"]:
        if hr_bytes < 1024.0:
            break
        if suffix != "TB":
            hr_bytes /= 1024.0

    if suffix == "B":
        return f"{int(hr_bytes)} {suffix}"
    return f"{hr_bytes:.2f} {suffix}"


def translated_property(text: str) -> property:
    return property(lambda self: gettext(text))  # type: ignore



from unittest import TestCase, skipIf

from eduvpn.nm import NMManager
from eduvpn.variants import EDUVPN
from tests.mock_config import mock_config


@skipIf(not NMManager(EDUVPN, None).available, "Network manager not available")
class TestNm(TestCase):
    def test_nm_available(self):
        nm_manager = NMManager(EDUVPN, None)
        nm_manager.available

    def test_import_ovpn(self):
        nm_manager = NMManager(EDUVPN, None)
        nm_manager.import_ovpn(mock_config)

    def test_get_add_connection(self):
        nm_manager = NMManager(EDUVPN, None)
        simple_connection = nm_manager.import_ovpn(mock_config)
        nm_manager.add_connection(simple_connection)

    def test_get_uuid(self):
        nm_manager = NMManager(EDUVPN, None)
        nm_manager.uuid



import json
from typing import Dict, List

from eduvpn_common.main import ServerType

from eduvpn.i18n import extract_translation


class DiscoOrganization:
    """The class that represents an organization from discovery
    :param: display_name: Dict[str, str]: The display name of the organizations
    :param: org_id: str: The organization ID
    """

    def __init__(
        self,
        display_name: Dict[str, str],
        org_id: str,
    ):
        self.display_name = display_name
        self.org_id = org_id

    @property
    def identifier(self):
        return self.org_id

    @property
    def category_id(self) -> ServerType:
        return ServerType.SECURE_INTERNET

    def __str__(self):
        return extract_translation(self.display_name)


class DiscoServer:
    """The class that represents a discovery server, this can be an institute access or secure internet server
    :param: base_url: str: The base URL of the server
    :param: display_name: Dict[str, str]: The display name of the server
    :param: server_type: str: The server type as a string
    """

    def __init__(
        self,
        base_url: str,
        display_name: Dict[str, str],
        server_type: str,
    ):
        self.base_url = base_url
        self.display_name = display_name
        self.server_type = server_type

    @property
    def identifier(self):
        return self.base_url

    @property
    def category_id(self) -> ServerType:
        return ServerType.INSTITUTE_ACCESS

    def __str__(self):
        return extract_translation(self.display_name)


def parse_disco_server(s: dict) -> DiscoServer:
    b_url = s["base_url"]
    display_name = s.get("display_name", "")
    # Mandatory
    server_type = s["server_type"]
    return DiscoServer(b_url, display_name, server_type)


def parse_disco_organization(o: dict) -> DiscoOrganization:
    display_name = o.get("display_name", "")
    org_id = o["org_id"]
    return DiscoOrganization(display_name, org_id)


def parse_disco_servers(json_str: str) -> List[DiscoServer]:
    d = json.loads(json_str)
    servers = d.get("server_list", [])
    disco_servers = []
    for s in servers:
        disco_server = parse_disco_server(s)
        disco_servers.append(disco_server)
    return disco_servers


def parse_disco_organizations(json_str: str) -> List[DiscoOrganization]:
    d = json.loads(json_str)
    organizations = d.get("organization_list", [])
    disco_orgs = []
    for o in organizations:
        disco_org = parse_disco_organization(o)
        disco_orgs.append(disco_org)
    return disco_orgs



import logging
from gettext import gettext as _
from gettext import ngettext

import gi

gi.require_version("Gtk", "3.0")
from eduvpn_common.main import EduVPN
from eduvpn_common.state import State, StateType
from gi.repository import GLib, Gio, Gtk
from gi.repository.Gio import ApplicationCommandLine

from eduvpn import i18n, notify
from eduvpn.app import Application
from eduvpn.settings import CONFIG_DIR_MODE
from eduvpn.ui.ui import EduVpnGtkWindow
from eduvpn.utils import init_logger, run_in_background_thread, ui_transition
from eduvpn.variants import ApplicationVariant

logger = logging.getLogger(__name__)


class EduVpnGtkApplication(Gtk.Application):
    def __init__(self, *args, app_variant: ApplicationVariant, common: EduVPN, **kwargs) -> None:
        super().__init__(  # type: ignore
            *args,
            application_id=app_variant.app_id,
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE,  # type: ignore
            **kwargs,
        )

        self.app = Application(app_variant, common)
        self.common = common
        self.common.register_class_callbacks(self)
        self.debug = False
        # Only allow a single window and track it on the app.
        self.window = None

        self.add_main_option(  # type: ignore
            "version",
            ord("v"),
            GLib.OptionFlags.NONE,  # type: ignore
            GLib.OptionArg.NONE,  # type: ignore
            "print version and exit",
            None,
        )
        self.add_main_option(  # type: ignore
            "debug",
            ord("d"),
            GLib.OptionFlags.NONE,  # type: ignore
            GLib.OptionArg.NONE,  # type: ignore
            "enable debug logging",
            None,
        )

    def do_startup(self) -> None:
        logger.debug("startup")
        Gtk.Application.do_startup(self)  # type: ignore
        i18n.initialize(self.app.variant)
        notify.initialize(self.app.variant)
        self.connection_notification = notify.Notification(self.app.variant)

    def do_shutdown(self) -> None:  # type: ignore
        logger.debug("shutdown")
        self.connection_notification.hide()
        Gtk.Application.do_shutdown(self)  # type: ignore

    def do_activate(self) -> None:
        logger.debug("activate")
        if not self.window:
            self.window = EduVpnGtkWindow(application=self)  # type: ignore
            self.window.present()  # type: ignore
            self.window.initialize()  # type: ignore
        else:
            self.window.on_reopen_window()

    def do_command_line(self, command_line: ApplicationCommandLine) -> int:  # type: ignore
        logger.debug(f"command line: {command_line}")
        options = command_line.get_options_dict()

        # unpack the commandline args into a dict
        options = options.end().unpack()

        if "version" in options:  # type: ignore
            # already handled
            return 0

        self.debug = "debug" in options  # type: ignore

        init_logger(self.debug, self.app.variant.logfile, CONFIG_DIR_MODE)

        self.activate()  # type: ignore
        return 0

    def on_quit(self, action: None = None, _param: None = None) -> None:
        logger.debug("quit")
        try:
            self.app.model.cancel()
            # Deregister the common library to save settings
            self.common.deregister()
        # Cleaning up is best effort
        except Exception as e:
            logger.debug("failed cleaning up library", e)

    def on_window_closed(self) -> None:
        logger.debug("window closed")
        self.on_quit()

    def enter_CopiedAnError(self):
        self.connection_notification.show(
            title=_("Error Copied"),
            message=_(
                "The error message has been copied to your clipboard. "
                "Report it at https://codeberg.org/eduVPN/linux-app if you think it is an issue."
            ),
        )

    def enter_SessionPendingExpiryState(self, h: int):
        hours_text = ngettext("{0} hour.", "{0} hours.", h).format(h)
        self.connection_notification.show(
            title=_("Connected"),
            message=_(f"Your session will expire in {hours_text} Renew the session to extend the validity."),
        )

    def enter_SessionExpiredState(self):
        self.connection_notification.show(
            title=_("Session expired"),
            message=_("Your session has expired. You have been disconnected from the VPN."),
        )

        @run_in_background_thread("expired-deactivate")
        def expired_deactivate():
            self.app.model.deactivate_connection(cleanup=False)

        expired_deactivate()

    @ui_transition(State.GOT_CONFIG, StateType.ENTER)
    def enter_NoActiveConnection(self, _old_state, _new_state):
        if not self.window.is_visible():
            # Quit the app if no window is open when the connection is deactivated.
            logger.debug("connection deactivated while window closed")
            self.on_quit()



from pathlib import Path

from eduvpn.utils import get_config_dir, get_prefix

prefix = get_prefix()

CONFIG_PREFIX = (Path(get_config_dir()).expanduser() / "eduvpn").resolve()
LETSCONNECT_CONFIG_PREFIX = (Path(get_config_dir()).expanduser() / "letsconnect").resolve()
CONFIG_DIR_MODE = 0o700  # Same as the Go library

CLIENT_ID = "org.eduvpn.app.linux"
LETSCONNECT_CLIENT_ID = "org.letsconnect-vpn.app.linux"
LANGUAGE = "en"
COUNTRY = "en-EN"

COUNTRY_MAP = Path(prefix + "/share/eduvpn/country_codes.json")
FLAG_PREFIX = prefix + "/share/eduvpn/images/flags/png/"
IMAGE_PREFIX = prefix + "/share/eduvpn/images/"
LC_IMAGE_PREFIX = prefix + "/share/letsconnect/images/"


EDUVPN_ICON = prefix + "/share/icons/hicolor/128x128/apps/org.eduvpn.client.png"
EDUVPN_NAME = "eduVPN"
EDUVPN_LOGO = IMAGE_PREFIX + "edu-vpn-logo.png"
EDUVPN_LOGO_DARK = IMAGE_PREFIX + "edu-vpn-logo-dark.png"
LETS_CONNECT_LOGO = LC_IMAGE_PREFIX + "letsconnect.png"
SERVER_ILLUSTRATION = LC_IMAGE_PREFIX + "server-illustration.png"
LETS_CONNECT_ICON = prefix + "/share/icons/hicolor/128x128/apps/org.letsconnect-vpn.client.png"
LETS_CONNECT_NAME = "Let's Connect!"
